/* 
 * Copyright (C) 2002, 2003 Emmanuel Saracco <esaracco@noos.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA. 
 */

#include "check.h"

#define UC_CHECK_IO_WHILE_CONDITION \
	( \
        	(!uc_check_gdk_input_end) && \
        	(!uc_check_cancel_get_value ()) && \
        	(!uc_check_ignore_item_get_value ()) && \
        	(!uc_check_timeout_get_value ()) && \
		(!uc_check_gdk_input_timeout ()) \
	)

#define UC_CHECK_WAIT_END_OTHER_CONDITIONS \
	( \
		(uc_check_cancel_get_value ()) || \
		(uc_check_ignore_item_get_value ()) \
	)


/*
 * struct to pass to gdk_input 
 * callbacks
 */
typedef struct _UCGetContentReadData
{
  guint uid;
  guint length;
}
UCGetContentReadData;

/*
 * struct containing path for "type"
 * images
 */
struct
{
  UCLinkType type;
  gchar *icon_file;
}
static uc_check_link_type_icon[] = {
  {LINK_TYPE_HREF, UC_PIXMAPS_DIR "/link_type_href.png"},
  {LINK_TYPE_FILE_HREF, UC_PIXMAPS_DIR "/link_type_href_local.png"},
  {LINK_TYPE_FILE_IMAGE, UC_PIXMAPS_DIR "/link_type_image_local.png"},
  {LINK_TYPE_EMAIL, UC_PIXMAPS_DIR "/link_type_mail.png"},
  {LINK_TYPE_IMAGE, UC_PIXMAPS_DIR "/link_type_image.png"},
  {LINK_TYPE_FRAME, UC_PIXMAPS_DIR "/link_type_frame.png"},
  {LINK_TYPE_NONE, UC_PIXMAPS_DIR "/link_type_unknown.png"}
};


/* global variable to know when a network operation is accomplished */
static gboolean uc_check_gdk_input_end = FALSE;

/* global variable to know when operation take to much
 * time */
static guint uc_check_gdk_input_timeout_value = 0;

/* TRUE if we are processing data in a gdk_input
 * method */
static gboolean uc_check_gdk_input_timeout_lock_value = FALSE;

/* unique id for a glist properties node */
static guint uc_check_node_uid = 1;

static gboolean uc_check_timeout_value = FALSE;

/* if TRUE, wait until it is FALSE */
static gboolean uc_check_wait_value = FALSE;

static gboolean uc_check_suspend_value = FALSE;

/* if FALSE, abort the check */
static gboolean uc_check_cancel_value = FALSE;

/* if FALSE, abort the check of current url*/
static gboolean uc_check_ignore_item_value = FALSE;

/* special glist to register timeout links in
 * -> at the end of the check this list will be added
 *  to the main glist links */
static GList *uc_check_list_timeout = NULL;

static void uc_check_right_frames_reset (void);
static void uc_check_insert_bad_link (void);
static GHashTable *uc_check_url_get_header_timeout (void);
static GHashTable *uc_check_url_get_header_ignored (void);
static gchar *uc_check_url_get_content (UCLinkProperties * prop,
					const gchar * current_path,
					const gchar * current_file,
					const gchar * current_port,
					const gchar * current_args,
					gchar * current_host);
static GHashTable *uc_check_url_get_header (UCLinkProperties * prop);
static gboolean uc_check_content_type_accepted (const gchar * content_type);
static gboolean uc_check_protocol_accepted (gchar * protocol);
static GdkPixbuf *uc_check_get_link_type_icon (const UCLinkType link_type);

static gboolean
uc_check_gdk_input_get_content_read (gpointer data, gint fd,
				     GdkInputCondition in);
static gboolean uc_check_gdk_input_get_header_read (gpointer data, gint fd,
						    GdkInputCondition in);
static gboolean uc_check_content_type_parsable (const gchar * content_type);
static gboolean uc_check_content_type_downloadable (const gchar *
						    content_type);
static gboolean uc_check_apply_filters (const UCLinkProperties * prop);
static gboolean uc_check_domain_accepted (gchar * domain);
static gboolean uc_check_directory_accepted (const gchar * directory);
static gboolean uc_check_chroot_authorized (const gchar * host,
					    const gchar * path);
static gboolean
uc_check_link_get_properties_proto_mailto (UCLinkProperties * prop);
static UCHTTPCheckReturn
uc_check_link_get_properties_proto_http (UCLinkProperties * prop,
					 const gchar * url,
					 const gchar * current_path,
					 gchar * current_host,
					 UCHTMLTag * tag);
static gboolean
uc_check_link_get_properties_proto_file (UCLinkProperties * prop,
					 const gchar * url,
					 const UCLinkType type);
static gboolean uc_check_wait_callback (gpointer data);
static gboolean uc_check_gdk_input_timeout (void);
static void uc_check_gdk_input_timeout_end (void);
static void uc_check_gdk_input_timeout_init (void);
static void uc_check_gdk_input_timeout_reset (void);
static void uc_check_gdk_input_timeout_lock (void);
static void uc_check_gdk_input_timeout_unlock (void);


/*
 * return TRUE if we consider that the given
 * status code is good
 */
gboolean
uc_check_status_is_good (const gchar * status)
{
  return ((status[0] == '2') ||
	  (!strcmp (status, STATUS_CODE_REDIRECTED)) ||
	  (!strcmp (status, STATUS_CODE_MOVED_PERMANENTLY)) ||
	  (!strcmp (status, STATUS_CODE_GOOD_EMAIL_LINK)));
}


/*
 * return TRUE if we consider that the given
 * status code is bad
 */
gboolean
uc_check_status_is_bad (const gchar * status)
{
  return ((status[0] != '2') && (status[0] != '9')
	  && (status[0] != '8')
	  && (strcmp (status, STATUS_CODE_REDIRECTED))
	  && (strcmp (status, STATUS_CODE_MOVED_PERMANENTLY))
	  && (strcmp (status, "503"))
	  && (strcmp (status, STATUS_CODE_TIMEOUT))
	  && (strcmp (status, STATUS_CODE_GOOD_EMAIL_LINK)));
}


/*
 * return TRUE if we consider that the given
 * status code is malformed
 */
gboolean
uc_check_status_is_malformed (const gchar * status)
{
  return ((!strcmp (status, STATUS_CODE_MALFORMED_LINK)) ||
	  (!strcmp (status, STATUS_CODE_FILE_PROTO_ERROR)));
}


/*
 * return TRUE if we consider that the given
 * status code is bad
 */
gboolean
uc_check_status_is_timeout (const gchar * status)
{
  return ((status[0] == '9' ||
	   !strcmp (status, "503") ||
	   !strcmp (status, STATUS_CODE_TIMEOUT))
	  &&
	  (strcmp (status, STATUS_CODE_IGNORED_LINK) &&
	   strcmp (status, STATUS_CODE_MALFORMED_LINK) &&
	   strcmp (status, STATUS_CODE_FILE_PROTO_ERROR)));
}


/*
 * return TRUE if the given status is the status
 * of a email link
 */
gboolean
uc_check_status_is_email (const gchar * status)
{
  return (!strcmp (status, STATUS_CODE_BAD_EMAIL_LINK) ||
	  !strcmp (status, STATUS_CODE_BAD_EMAIL_LINK_MX) ||
	  !strcmp (status, STATUS_CODE_GOOD_EMAIL_LINK));
}


/*
 * return TRUE if the link has been ignored
 * by the user
 */
gboolean
uc_check_status_is_ignored (const gchar * status)
{
  return (!strcmp (status, STATUS_CODE_IGNORED_LINK));
}


/*
 * set the timeout flag
 */
void
uc_check_timeout_set_value (const gboolean value)
{
  uc_check_timeout_value = value;
}


/*
 * get the timeout flag
 */
gboolean
uc_check_timeout_get_value (void)
{
  return uc_check_timeout_value;
}


/*
 * return the label for a given link type
 */
static GdkPixbuf *
uc_check_get_link_type_icon (const UCLinkType link_type)
{
  guint i = 0;

  for (i = 0; i < LINK_TYPES_LEN; i++)
    if (uc_check_link_type_icon[i].type == link_type)
      return gdk_pixbuf_new_from_file
	(uc_check_link_type_icon[i].icon_file, NULL);

  return gdk_pixbuf_new_from_file (UC_PIXMAPS_DIR "/list_unknown.png", NULL);
}


/*
 * display recursively urls in the treeview
 */
void
uc_check_display_list (GList * list, GtkTreeIter iter[], guint i_iter,
		       gchar * path)
{
  UCLinkProperties *prop = NULL;
  GList *item = NULL;
  guint pos = 0;
  gchar no_value[] = " ";


  item = list;
  while (item)
    {
      gchar *value = NULL;
      gchar *label = NULL;

      prop = (UCLinkProperties *) item->data;

      item = g_list_next (item);

      if (!prop)
	{
	  g_warning ("prop is NULL!");

	  continue;
	}

      /*
       * we do not display meta links nor nodes to remove from
       * bookmarks */
      if (prop->link_type == LINK_TYPE_META || prop->to_delete)
	continue;

      /* here we build the treeview path to automatically select
       * a row according to the search result selection */
      g_free (prop->treeview_path);
      prop->treeview_path = (!path) ?
	g_strdup_printf ("%u", pos) : g_strdup_printf ("%s:%u", path, pos);

      value = (prop->link_value) ? prop->link_value : no_value;
      label = (prop->label) ? prop->label : no_value;

      if (!i_iter)
	{
	  gtk_tree_store_append (treestore, &iter[i_iter], NULL);
	  gtk_tree_store_set (treestore, &iter[0],
			      UID_COLUMN, prop->uid,
			      LINK_STATUS_ICON_COLUMN,
			      prop->status_icon,
			      LINK_ICON_COLUMN,
			      prop->link_icon, LABEL_COLUMN,
			      label, URL_COLUMN, value, -1);
	}
      else
	{
	  gtk_tree_store_append (treestore, &iter[i_iter], &iter[i_iter - 1]);
	  gtk_tree_store_set (treestore, &iter[i_iter],
			      UID_COLUMN, prop->uid,
			      LINK_STATUS_ICON_COLUMN,
			      prop->status_icon,
			      LINK_ICON_COLUMN,
			      prop->link_icon, LABEL_COLUMN,
			      label, URL_COLUMN, value, -1);
	}

      if (prop->childs)
	uc_check_display_list (prop->childs, iter, i_iter + 1,
			       prop->treeview_path);

      pos++;
    }
}

/*
 * display url list with a mask
 */
void
uc_check_display_list_with_only (GList * list, GtkTreeIter iter,
				 const guint view_type)
{
  GList *item = NULL;
  UCLinkProperties *prop = NULL;
  gchar *status_code = NULL;
  gboolean display = FALSE;

  item = list;
  while (item)
    {
      prop = (UCLinkProperties *) item->data;

      item = g_list_next (item);

      /*
       * we do not want to display nodes to remove from
       * bookmarks */
      if (prop->to_delete)
	continue;

      if (prop->link_type != LINK_TYPE_BOOKMARK_FOLDER)
	{
	  status_code = g_hash_table_lookup (prop->header, "Status");

	  switch (view_type)
	    {
	    case UC_CHECK_DISPLAY_BADLINKS:
	      display = uc_check_status_is_bad (status_code);
	      break;

	    case UC_CHECK_DISPLAY_GOODLINKS:
	      display = uc_check_status_is_good (status_code);
	      break;

	    case UC_CHECK_DISPLAY_TIMEOUTS:
	      display = uc_check_status_is_timeout (status_code);
	      break;

	    case UC_CHECK_DISPLAY_MALFORMED:
	      display = uc_check_status_is_malformed (status_code);
	      break;

	    }

	  if (display)
	    {
	      gtk_tree_store_append (treestore, &iter, NULL);
	      gtk_tree_store_set (treestore, &iter,
				  UID_COLUMN, prop->uid,
				  LINK_STATUS_ICON_COLUMN,
				  prop->status_icon,
				  LINK_ICON_COLUMN,
				  prop->link_icon, LABEL_COLUMN,
				  prop->label, URL_COLUMN, prop->link_value,
				  -1);
	    }
	}

      if (prop->childs)
	uc_check_display_list_with_only (prop->childs, iter, view_type);
    }
}

/*
 * return the id of the current selected row in the url
 * treeview
 */
gint
uc_check_treeview_get_selected_row_id (void)
{
  GtkTreeIter iter;
  GtkTreeModel *model = NULL;
  GtkTreeSelection *select = NULL;
  gint ret = -1;

  if (uc_lists_checked_links_is_empty () ||
      !(select = gtk_tree_view_get_selection (treeview)))
    return ret;

  if (gtk_tree_selection_get_selected (select, &model, &iter))
    gtk_tree_model_get (model, &iter, 0, &ret, -1);

  return ret;
}


/*
 * get the source of a given page
 */
static gchar *
uc_check_url_get_content (UCLinkProperties * prop,
			  const gchar * current_path,
			  const gchar * current_file,
			  const gchar * current_port,
			  const gchar * current_args, gchar * current_host)
{
  UCGetContentReadData read_data;
  gchar buffer[BUFFER_LEN + 1] = { 0 };
  guint sock = 0;
  guint gifd = 0;
  gchar *h = NULL;

  h = (strlen (prop->h_name) ? prop->h_name : current_host);

  uc_application_display_state_message (CHECK_MESSAGE_LABEL_THIRD,
					_
					("Retrieving source (<b>0</b> byte)..."));
  uc_application_update_ui ();

  if (!(sock = uc_connection_open (h, atoi (prop->port))))
    return NULL;

  if (uc_project_get_use_proxy ())
    snprintf (buffer, BUFFER_LEN, "GET %s://%s%s%s%s%s HTTP/1.0\r\n"
	      "User-Agent: %s\r\n"
	      "Host: %s\r\n"
	      "Referer: %s\r\n"
	      "Accept: */*\r\n\r\n",
	      prop->protocol,
	      h,
	      (prop->path[0] != '/' && prop->file[0] != '/') ? "/" : "",
	      prop->path, prop->file, prop->args, USER_AGENT, h, REFERER);
  else
    snprintf (buffer, BUFFER_LEN, "GET %s%s%s%s HTTP/1.0\r\n"
	      "User-Agent: %s\r\n"
	      "Host: %s\r\n"
	      "Referer: %s\r\n"
	      "Accept: */*\r\n\r\n",
	      (prop->path[0] != '/' && prop->file[0] != '/') ? "/" : "",
	      prop->path, prop->file, prop->args, USER_AGENT, h, REFERER);

  write (sock, buffer, strlen (buffer));

  uc_check_gdk_input_end = FALSE;
  read_data.uid = prop->uid;
  read_data.length = 0;
  gifd = gdk_input_add (sock, GDK_INPUT_READ,
			(GdkInputFunction)
			uc_check_gdk_input_get_content_read, &read_data);

  uc_check_gdk_input_timeout_init ();
  while (UC_CHECK_IO_WHILE_CONDITION)
    uc_application_update_ui ();
  gdk_input_remove (gifd);
  uc_check_gdk_input_timeout_end ();

  /* quit if a timeout occured */
  if (uc_check_timeout_get_value ())
    {
      close (sock);

      return NULL;
    }

  if (!g_hash_table_lookup (prop->header, "Content-Length"))
    g_hash_table_insert (prop->header, g_strdup (CONTENT_LENGTH),
			 g_strdup_printf ("%u",
					  read_data.length -
					  prop->header_size));

  close (sock);

  return uc_cache_get_source (prop->uid);
}


/*
 * set a timeout header
 */
static GHashTable *
uc_check_url_get_header_timeout (void)
{
  GHashTable *tmp = NULL;

  if (uc_check_timeout_get_value ())
    return NULL;

  tmp = g_hash_table_new (g_str_hash, g_str_equal);
  g_hash_table_insert (tmp, g_strdup ("Status"),
		       g_strdup (STATUS_CODE_NO_DATA));
  g_hash_table_insert (tmp, g_strdup (CONTENT_TYPE),
		       g_strdup (CONTENT_TYPE_HTML));

  return tmp;
}


/*
 * set a item ignored header
 */
static GHashTable *
uc_check_url_get_header_ignored (void)
{
  GHashTable *tmp = NULL;

  tmp = g_hash_table_new (g_str_hash, g_str_equal);
  g_hash_table_insert (tmp, g_strdup ("Status"),
		       g_strdup (STATUS_CODE_IGNORED_LINK));
  g_hash_table_insert (tmp, g_strdup (CONTENT_TYPE),
		       g_strdup (CONTENT_TYPE_HTML));

  uc_report_set_ignoredlinks (0);
  uc_check_ignore_item_set_value (FALSE);

  return tmp;
}


/*
 * set the global variable who determine if we must
 * suspend or continue the check
 */
void
uc_check_suspend_continue (void)
{
  GtkWidget *widget = NULL;

  uc_check_suspend_value = (!uc_check_suspend_get_value ());

  UC_GLADE_GET_WIDGET ("btSuspend_progressDialog", glade_progress_dialog,
		       widget);

  if (uc_check_suspend_get_value ())
    {
      UC_WIDGET_SET_SENSITIVE (glade_progress_dialog,
			       "btIgnore_progressDialog", FALSE);
      gtk_button_set_label (GTK_BUTTON (widget), _("Continue"));
    }
  else
    {
      UC_WIDGET_SET_SENSITIVE (glade_progress_dialog,
			       "btIgnore_progressDialog", TRUE);
      gtk_button_set_label (GTK_BUTTON (widget), _("Suspend"));
    }
}


/*
 * read the page content of the given descriptor
 */
static gboolean
uc_check_gdk_input_get_content_read (gpointer buffer, gint fd,
				     GdkInputCondition in)
{
  UCGetContentReadData *read_data = (UCGetContentReadData *) buffer;
  gint nbr_read = -1;
  gchar buffer_recv[BUFFER_LEN + 1] = { 0 };
  gchar msg[BUFFER_URL_LEN + 1] = { 0 };
  FILE *fdf;
  gchar *path = NULL;
  gboolean ret = TRUE;

  if (!fd)
    {
      uc_check_timeout_set_value (TRUE);
      uc_check_gdk_input_end = TRUE;

      return FALSE;
    }

  uc_check_gdk_input_timeout_reset ();
  uc_check_gdk_input_timeout_lock ();

  nbr_read = read (fd, buffer_recv, BUFFER_LEN);

  if (nbr_read > 0)
    {
      read_data->length += nbr_read;

      snprintf (msg, BUFFER_URL_LEN,
		_("Retrieving source (<b>%u</b> bytes)..."),
		read_data->length);
      uc_application_display_state_message (CHECK_MESSAGE_LABEL_THIRD, msg);
      uc_application_update_ui ();

      path = uc_utils_convert_uid2file (read_data->uid);
      fdf = fopen (path, "a");
      fwrite (buffer_recv, 1, nbr_read, fdf);
      fclose (fdf);
      g_free (path);
    }
  else if (nbr_read <= 0)
    {
      if (nbr_read < 0)
	uc_check_timeout_set_value (TRUE);

      uc_check_gdk_input_end = TRUE;
      ret = FALSE;
    }

  uc_check_gdk_input_timeout_unlock ();
  uc_application_update_ui ();

  return ret;
}


/*
 * read the page header on the given descriptor
 */
static gboolean
uc_check_gdk_input_get_header_read (gpointer data, gint fd,
				    GdkInputCondition in)
{
  gint nbr_read = -1;

  if (!fd)
    return FALSE;

  uc_check_gdk_input_timeout_reset ();
  uc_check_gdk_input_timeout_lock ();

  nbr_read = read (fd, (gchar *) data, BUFFER_LEN);

  if (nbr_read < 0)
    uc_check_timeout_set_value (TRUE);

  uc_check_gdk_input_end = TRUE;

  uc_check_gdk_input_timeout_unlock ();
  uc_application_update_ui ();

  return FALSE;
}


/*
 * put all the header's informations in a
 * hashtable
 */
static GHashTable *
uc_check_url_get_header (UCLinkProperties * prop)
{
  GHashTable *tmp = NULL;
  gchar **lines = NULL;
  gchar *host = prop->h_name;
  gchar *path = prop->path;
  gchar *file = prop->file;
  gchar *args = prop->args;
  gchar *port = prop->port;
  gchar buffer[BUFFER_LEN + 1] = { 0 };
  gchar url[BUFFER_URL_LEN + 1] = { 0 };
  gchar buffer_recv[BUFFER_LEN + 1] = { 0 };
  gchar str[HEADER_STATUS_CODE_LEN + 1] = { 0 };
  guint i = 0;
  guint sock = 0;
  gint gifd = 0;

  uc_application_display_state_message
    (CHECK_MESSAGE_LABEL_THIRD, _("Retrieving header..."));
  uc_application_update_ui ();


  if (!(sock = uc_connection_open (host, atoi (port))))
    {
      if (!uc_check_timeout_get_value ())
	uc_check_alarm_callback (0);

      return NULL;
    }

  if (uc_project_get_use_proxy ())
    snprintf (url, BUFFER_URL_LEN, "%s://%s%s%s%s%s",
	      g_strdown (prop->protocol), host,
	      (path[0] != '/'
	       && file[0] != '/') ? "/" : "", path, file, args);
  else
    snprintf (url, BUFFER_URL_LEN, "%s%s%s%s",
	      (path[0] != '/'
	       && file[0] != '/') ? "/" : "", path, file, args);

  snprintf (buffer, BUFFER_LEN,
	    "HEAD %s HTTP/1.0\r\n"
	    "User-Agent: %s\r\n"
	    "Host: %s\r\n" "\r\n", url, USER_AGENT, host);

  write (sock, buffer, strlen (buffer));

  /* process the reading in background */
  uc_check_gdk_input_end = FALSE;
  gifd = gdk_input_add (sock, GDK_INPUT_READ,
			(GdkInputFunction)
			uc_check_gdk_input_get_header_read, &buffer_recv);

  /* wait for the end of the reading operation */
  uc_check_gdk_input_timeout_init ();
  while (UC_CHECK_IO_WHILE_CONDITION)
    uc_application_update_ui ();
  gdk_input_remove (gifd);
  uc_check_gdk_input_timeout_end ();

  /* quit if a timeout occured */
  if (uc_check_timeout_get_value ())
    {
      close (sock);

      return NULL;
    }

  if (!strlen (buffer_recv))
    return (uc_check_ignore_item_get_value ())?
      uc_check_url_get_header_ignored () : uc_check_url_get_header_timeout ();

  memcpy (&str, &buffer_recv[HEADER_STATUS_CODE_FIRST_POS],
	  HEADER_STATUS_CODE_LEN);
  tmp = g_hash_table_new (g_str_hash, g_str_equal);
  g_hash_table_insert (tmp, g_strdup ("Status"), g_strdup (str));

  prop->header_size = strlen (buffer_recv);
  i = 0;
  lines = g_strsplit (buffer_recv, "\r\n", 0);
  while (lines[i])
    {
      gchar **key_value = g_strsplit (lines[i], ":", 2);

      if (key_value[0] && key_value[1])
	{
	  uc_utils_strstrip (key_value[0]);
	  uc_utils_strstrip (key_value[1]);

	  /* FIXME: there is some problems on case with some
	   * header fields */
	  g_hash_table_insert (tmp,
			       (!strcmp (key_value[0], "Content-type")) ?
			       g_strdup (CONTENT_TYPE) :
			       g_strdup (key_value[0]),
			       g_strdup (key_value[1]));
	}

      g_strfreev (key_value);
      i++;
    }
  g_strfreev (lines);

  close (sock);

  return tmp;
}


/*
 * return the item corresponding to teh given URL
 * id found in the already checked list
 */
static UCLinkProperties *
uc_check_link_already_checked (gchar * url)
{
  return (UCLinkProperties *) uc_lists_checked_links_lookup (url);
}


/*
 * check if the current link have been already
 * checked, and register it if it was not
 */
gboolean
uc_check_link_already_checked_with_insert (UCLinkProperties * prop,
					   gchar * url)
{
  UCLinkProperties *prop_tmp = NULL;
  gboolean already_checked = FALSE;

  prop_tmp = uc_check_link_already_checked (url);

  /* if we are checking bookmarks, do not check for already
   * checked links */
  if (uc_project_get_bookmarks_file ())
    return FALSE;

  if ((already_checked = (prop_tmp != NULL)) &&
      (!uc_lists_similar_links_exist
       (prop_tmp->similar_links_parents, prop->url)))
    uc_lists_similar_links_append ((gpointer) prop_tmp, (gpointer) prop->url);

  if (already_checked)
    {
      uc_application_display_state_message (CHECK_MESSAGE_LABEL_THIRD,
					    _("Item already checked..."));
      uc_application_update_ui ();

      uc_utils_debug ("        (already checked) %s\n", url);
    }

  return already_checked;
}


/*
 * insert a malformed item in the
 * url list
 */
void
uc_check_insert_malformed_link (const UCHTMLTag * tag, GList * list)
{
  UCLinkProperties *prop = NULL;
  UCStatusCode *sc;

  prop = uc_check_link_properties_node_new ();
  prop->uid = uc_check_node_uid++;
  prop->parent = currentItem.parent;
  prop->link_icon = uc_check_get_link_type_icon (tag->type);
  sc = uc_application_get_status_code_properties (STATUS_CODE_MALFORMED_LINK);
  prop->status_icon = sc->icon_file;
  prop->link_value = g_strdup ("");
  prop->label = g_strdup (tag->label);
  prop->protocol = g_strdup ("");
  prop->h_name = g_strdup ("");
  prop->domain = g_strdup ("");
  prop->port = g_strdup (URL_DEFAULT_PORT);
  prop->path = g_strdup ("");
  prop->file = g_strdup ("");
  prop->args = g_strdup ("");
  prop->url = g_strdup ("");
  prop->header = g_hash_table_new (g_str_hash, g_str_equal);
  g_hash_table_insert (prop->header, g_strdup ("Status"),
		       g_strdup (STATUS_CODE_MALFORMED_LINK));

  uc_report_set_malformedlinks (0);

  uc_utils_debug ("        (malformed) <no url>\n");

  list = g_list_append (list, (gpointer) prop);
  uc_lists_already_checked_links_append (prop->url, (gpointer) prop);
}


/*
 * register a link in the checked urls
 * list
 */
GList *
uc_check_register_link (const gchar * normalized_url, UCLinkProperties * lp)
{
  gchar *status = NULL;

  /* append link to the already checked temporary 
   * list */
  uc_lists_already_checked_links_append (normalized_url, (gpointer) lp);

  /* we do not register bookmark's foldres */
  if (lp->link_type != LINK_TYPE_BOOKMARK_FOLDER)
    {
      status = g_hash_table_lookup (lp->header, "Status");

      if (!status)
	g_warning ("-> type: %u\nurl: %s\nlabel: %s\n", lp->link_type,
		   lp->url, lp->label);

      /* timeouts */
      if (uc_check_status_is_timeout (status))
	uc_report_set_timedoutlinks (0);

      /* good links */
      else if (uc_check_status_is_good (status))
	uc_report_set_goodlinks (0);

      /* bad links */
      else if (uc_check_status_is_bad (status))
	uc_report_set_badlinks (0);

      /* malformed link */
      else if (uc_check_status_is_malformed (status))
	uc_report_set_malformedlinks (0);

      /* display debug informations if requested */
      uc_utils_debug ("        (%s) %s\n", status, normalized_url);
    }

  uc_utils_debug_dump_link_properties (lp);

  return (gpointer) lp;
}


/*
 * test if the MX of the selected email address link is valid.
 * launch a popup to inform the user of its state
 */
void
uc_check_mx_is_valid (void)
{
  UCLinkProperties *prop = NULL;
  gchar *mx = NULL;
  guint id = 0;

  if ((id = uc_check_treeview_get_selected_row_id ()) < 0)
    {
      uc_application_dialog_show (_
				  ("Please, select a item first."),
				  GTK_MESSAGE_WARNING);
      return;
    }

  prop = uc_lists_checked_links_lookup_by_uid (id);

  if (prop->domain)
    {
      gchar *msg = NULL;

      mx = uc_utils_get_mx (prop->domain);
      if (uc_utils_mx_is_valid (mx))
	{
	  msg = g_strdup_printf (_("MX host %s for the specified "
				   "domain is alive and listen on "
				   "port 25."), mx);
	  uc_application_dialog_show (msg, GTK_MESSAGE_INFO);
	}
      else
	uc_application_dialog_show (_("MX host does not respond or "
				      "the domain of this email is not valid."),
				    GTK_MESSAGE_WARNING);
    }
  else
    uc_application_dialog_show (_("The domain of this email is not valid."),
				GTK_MESSAGE_WARNING);

  g_free (mx);
}


/*
 * add timeout links to the main
 * urls list
 */
void
uc_check_register_timeout_links (GList ** list)
{
  GList *item = NULL;

  item = uc_check_list_timeout;
  while (item)
    {
      *list = g_list_append (*list, (gpointer) item->data);
      item = g_list_next (item);
    }
}


/*
 * return TRUE is there is no filter for the given 
 * directory
 */
static gboolean
uc_check_directory_accepted (const gchar * path)
{
  gchar **items = NULL;
  gchar *tmp = NULL;
  guint size = 0;
  guint i = 0;
  gboolean found = FALSE;

  /* root can not be filtered */
  if (!strcmp (path, "/"))
    return TRUE;

  tmp = (path[strlen (path) - 1] != '/') ?
    g_strconcat (path, "/", NULL) : g_strdup (path);

  UC_CONFIG_GET_VECTOR
    (UC_CONFIG_PROPERTIES_FILTERS_REJECT_DIRECTORIES, &items, &size);
  if (items && strlen (items[0]))
    while ((i < size) && (!(found = (strstr (tmp, items[i++]) != NULL))));
  g_strfreev (items);

  g_free (tmp);

  return (!found);
}


/*
 * return TRUE is the domain there is no filter for 
 * the given hostname
 */
static gboolean
uc_check_domain_accepted (gchar * domain)
{
  gchar **items = NULL;
  guint size = 0;
  guint i = 0;
  gchar *tmp = NULL;
  gboolean found = FALSE;

  /* checked host is not filtered */
  if (!strcmp (uc_project_get_current_host (), domain))
    return TRUE;

  /* extract the domain of the complete hostname */
  tmp = (strchr (domain, '.') != strrchr (domain, '.')) ?
    strchr (domain, '.') + 1 : domain;

  UC_CONFIG_GET_VECTOR (UC_CONFIG_PROPERTIES_FILTERS_REJECT_DOMAINS,
			&items, &size);
  if (items && strlen (items[0]))
    while ((i < size) && (!(found = (!strcmp (tmp, items[i++])))));
  g_strfreev (items);

  return (!found);
}


/*
 * return TRUE if the given path is the check has been
 * restricted to the project directory and its subdirectories and
 * if the given path is in the current or in the subdirectories
 */
static gboolean
uc_check_chroot_authorized (const gchar * host, const gchar * path)
{
  gboolean ret = TRUE;
  gboolean same_host = FALSE;
  gboolean same_path = FALSE;

  same_host = uc_utils_memcasecmp (host, uc_project_get_current_host ());
  same_path = uc_utils_memcasecmp (path, uc_project_get_chroot_path ());

  if (!uc_project_get_limit_local ())
    {
      if (uc_project_get_check_chroot () && same_host)
	ret = same_path;
    }
  else if (!uc_project_get_check_chroot ())
    ret = same_host;
  else
    ret = (same_host && same_path);

  return ret;
}


/*
 * check if the given url is ok with
 * the filters
 */
static gboolean
uc_check_apply_filters (const UCLinkProperties * prop)
{
  gchar *content_type = NULL;

  return (prop &&
	  prop->header &&
	  (!strcmp (prop->protocol, "MAILTO") ||
	   ((content_type =
	     g_hash_table_lookup (prop->header, CONTENT_TYPE)) &&
	    uc_check_content_type_accepted (content_type))) &&
	  uc_check_directory_accepted (prop->url) &&
	  uc_check_domain_accepted (prop->h_name));
}


/*
 * increment the internal timeout counter
 */
void
uc_check_gdk_input_timeout_increment (void)
{
  uc_check_gdk_input_timeout_value++;
}


/*
 * avoid timeout when manipulating data
 */
static void
uc_check_gdk_input_timeout_lock (void)
{
  uc_check_gdk_input_timeout_lock_value = TRUE;
}


/*
 * authorize timeout
 */
static void
uc_check_gdk_input_timeout_unlock (void)
{
  uc_check_gdk_input_timeout_lock_value = FALSE;
}


/*
 * return TRUE if we can increment timeout 
 * value
 */
gboolean
uc_check_gdk_input_timeout_locked (void)
{
  return uc_check_gdk_input_timeout_lock_value;
}


/*
 * initialize the timeout flag for
 * gdk_input methods
 */
static void
uc_check_gdk_input_timeout_init (void)
{
  uc_check_gdk_input_timeout_value = 0;
  uc_check_gdk_input_timeout_unlock ();
}


/*
 * TRUE if the waiting time has
 * been passed
 */
static gboolean
uc_check_gdk_input_timeout (void)
{
  return (uc_check_gdk_input_timeout_value >
	  uc_project_get_check_timeout () + 1);
}


/*
 * end the timeout for gdk input methods
 * -> if a timeout occured, set the main
 *  timeout flag to TRUE
 */
static void
uc_check_gdk_input_timeout_end (void)
{
  if (uc_check_gdk_input_timeout_value >= uc_project_get_check_timeout ())
    uc_check_alarm_callback (0);

  uc_check_gdk_input_timeout_value = 0;
}

/*
 * same as init () method
 */
static void
uc_check_gdk_input_timeout_reset (void)
{
  uc_check_gdk_input_timeout_init ();
}


/*
 * set a SIGALRM
 */
void
uc_check_alarm_add (const guint seconds, gpointer callback)
{
  struct sigaction action;

  uc_check_timeout_set_value (FALSE);
  sigemptyset (&(action.sa_mask));
  action.sa_flags = 0;
  action.sa_handler = callback;

  sigaction (SIGALRM, &action, NULL);
  alarm (seconds);
}


/*
 * remove a SIGALRM
 */
void
uc_check_alarm_remove (void)
{
  alarm (0);
}


/*
 * set the cancel check flag.
 * this flag is tested during the check to know
 * if the check must aborting or continuing
 */
void
uc_check_cancel_set_value (const gboolean value)
{
  if (value && !cancel_dialog && !GTK_IS_WIDGET (cancel_dialog))
    {
      GdkColor color;

      UC_GLADE_INIT ("cancel_dialog", glade_cancel_dialog, cancel_dialog);

      gdk_color_parse ("green", &color);
      gtk_widget_modify_bg (cancel_dialog, GTK_STATE_NORMAL, &color);

      gtk_widget_show_all (cancel_dialog);
    }
  else if (cancel_dialog)
    {
      gtk_widget_destroy (cancel_dialog);
      cancel_dialog = NULL;
    }

  uc_check_cancel_value = value;
}


/*
 * return TRUE is the check is to suspend and
 * FALSE otherwise
 */
gboolean
uc_check_suspend_get_value (void)
{
  return uc_check_suspend_value;
}


/*
 * return TRUE is the check is to cancel and
 * FALSE otherwise
 */
gboolean
uc_check_cancel_get_value (void)
{
  return uc_check_cancel_value;
}


/*
 * return TRUE is the current item must be
 * ignored
 */
gboolean
uc_check_ignore_item_get_value (void)
{
  return uc_check_ignore_item_value;
}


/*
 * set ignore item value 
 */
void
uc_check_ignore_item_set_value (const gboolean value)
{
  uc_check_ignore_item_value = value;
}


/*
 * reset all items in the right frames
 */
static void
uc_check_right_frames_reset (void)
{
  GtkWidget *widget = NULL;

  UC_GLADE_GET_WIDGET ("label_informations_message", glade_main_window,
		       widget);
  if (widget)
    gtk_label_set_text (GTK_LABEL (widget), "");

  UC_GLADE_GET_WIDGET ("text_informations", glade_main_window, widget);
  if (widget)
    gtk_text_buffer_set_text (gtk_text_view_get_buffer
			      (GTK_TEXT_VIEW (widget)), "", -1);

  uc_check_report_reset ();
}


/*
 * free the timeout links list
 */
void
uc_check_timeout_links_free (void)
{
  g_list_free (uc_check_list_timeout);
  uc_check_list_timeout = NULL;
}


/*
 * this function must be called before
 * any check, to initialize some elements
 */
void
uc_check_reset (void)
{
  gchar *title = NULL;

  // FIXME
  //if (treestore)
  //  g_object_unref (treestore);
  uc_application_remove_paths ();
  uc_application_make_paths ();
  uc_lists_already_checked_free ();
  uc_check_timeout_links_free ();
  uc_check_right_frames_reset ();

  uc_refresh_page = FALSE;
  uc_check_cancel_set_value (FALSE);
  uc_check_suspend_value = FALSE;
  uc_check_ignore_item_set_value (FALSE);
  uc_check_node_uid = 1;

  title = g_strconcat ("gURLChecker v", VERSION, NULL);
  gtk_window_set_title (GTK_WINDOW (main_window), title);
  g_free (title);
}


/*
 * the callback to pass to uc_check_alarm_add()
 * function
 */
void
uc_check_alarm_callback (int dum)
{
  uc_check_timeout_set_value (TRUE);

  if (!uc_project_get_check_is_current () ||
      (currentItem.url && uc_utils_memcasecmp (currentItem.url, "mailto:")))
    return;

  if (!uc_lists_checked_links_exist (currentItem.url))
    uc_check_insert_bad_link ();

  uc_report_display_update ();
}


/*
 * insert a bad item in the url list (typically
 * timeout or bad host)
 */
static void
uc_check_insert_bad_link (void)
{
  gchar *url = currentItem.url;
  gchar *hostname = NULL;
  UCLinkProperties *prop = NULL;
  UCStatusCode *sc;

  if (!url)
    url = uc_project_get_url ();

  prop = uc_check_link_properties_node_new ();
  prop->uid = uc_check_node_uid++;
  prop->xml_node = currentItem.xml_node;
  prop->parent = currentItem.parent;
  prop->link_icon = uc_check_get_link_type_icon (LINK_TYPE_NONE);
  sc = uc_application_get_status_code_properties (STATUS_CODE_TIMEOUT);
  prop->status_icon = sc->icon_file;
  prop->link_value = g_convert (url, -1, "UTF-8",
				uc_project_get_local_charset (),
				NULL, NULL, NULL);
  prop->label = g_strdup ((currentItem.label) ? currentItem.label : "");
  prop->protocol = uc_url_get_protocol (url);
  prop->h_name = g_strdup ("");
  prop->domain = g_strdup ("");
  prop->port = g_strdup (URL_DEFAULT_PORT);
  prop->path = g_strdup ("");
  prop->file = g_strdup ("");
  prop->args = g_strdup ("");
  prop->header = g_hash_table_new (g_str_hash, g_str_equal);

  hostname = uc_url_get_hostname ("", url);
  if (!strlen (hostname))
    prop->url = uc_url_normalize (uc_project_get_current_host (), "", url);
  else
    prop->url = uc_url_normalize ("", "", url);
  g_free (hostname);

  g_hash_table_insert (prop->header, g_strdup ("Status"),
		       g_strdup (STATUS_CODE_TIMEOUT));

  uc_report_set_timedoutlinks (0);

  uc_check_list_timeout =
    g_list_append (uc_check_list_timeout, (gpointer) prop);
  uc_lists_already_checked_links_append (prop->url, (gpointer) prop);

  uc_utils_debug ("        (timeout) %s\n", prop->url);
}


/*
 * return TRUE if the given content-type is parsable
 * (if it could contain links)
 */
static gboolean
uc_check_content_type_parsable (const gchar * content_type)
{
  return ((content_type != NULL)
	  && ((strstr (content_type, "htm") != NULL)
	      || (strstr (content_type, "php") != NULL)));
}


/*
 * return TRUE if the given content-type is dowloadable
 */
static gboolean
uc_check_content_type_downloadable (const gchar * content_type)
{
  return
    ((content_type != NULL) &&
     (((uc_project_get_download_images_content ()) &&
       (strstr (content_type, "image") != NULL)) ||
      ((!uc_check_content_type_parsable (content_type)) &&
       (uc_project_get_download_archives_content ()))));
}


/*
 * return TRUE if the given content-type is authorized
 * to be checked
 */
static gboolean
uc_check_content_type_accepted (const gchar * content_type)
{
  gchar **documents;
  gchar **images;
  guint documents_size = 0;
  guint images_size = 0;
  guint i = 0;

  if (!content_type)
    return FALSE;

  UC_CONFIG_GET_VECTOR
    (UC_CONFIG_PROPERTIES_URL_CONTENT_TYPES_DOCUMENTS, &documents,
     &documents_size);
  UC_CONFIG_GET_VECTOR (UC_CONFIG_PROPERTIES_URL_CONTENT_TYPES_IMAGES,
			&images, &images_size);

  if (!strlen (documents[0]))
    documents_size = 0;
  if (!strlen (images[0]))
    images_size = 0;

  for (i = 0; i < documents_size; i++)
    if (strstr (content_type, documents[i]))
      {
	g_strfreev (images);
	g_strfreev (documents);
	return FALSE;
      }
  g_strfreev (documents);

  for (i = 0; i < images_size; i++)
    if (strstr (content_type, images[i]))
      {
	g_strfreev (images);
	return FALSE;
      }
  g_strfreev (images);

  return TRUE;
}


/*
 * return TRUE if the given protocol is authorized
 * to be checked.
 */
static gboolean
uc_check_protocol_accepted (gchar * protocol)
{
  gchar **accepted = UC_CHECK_PROTOCOLS_GET_ACCEPTED;
  guint i = 0;
  gboolean found = FALSE;

  g_strup (protocol);

  while ((i < UC_CHECK_PROTOCOLS_ACCEPTED_LEN) &&
	 (!(found = (!strcmp (accepted[i++], protocol)))));

  g_strfreev (accepted);

  /* FIXME */
  if ((!strcmp (protocol, "MAILTO")) && (uc_project_get_proto_mailto ()))
    found = TRUE;
  else if ((!strcmp (protocol, "FILE")) &&
	   ((uc_project_get_proto_file_is_error ())
	    || (uc_project_get_proto_file_check ())))
    found = TRUE;

  return found;
}


/*
 * callback for uc_check_wait ()
 */
static gboolean
uc_check_wait_callback (gpointer data)
{
  uc_check_wait_value = FALSE;
  return TRUE;
}


/*
 * wait
 */
void
uc_check_wait (void)
{
  gchar buffer[BUFFER_LEN + 1] = { 0 };
  guint32 id = 0;

  if (!uc_project_get_check_wait ())
    return;

  snprintf (buffer, BUFFER_LEN,
	    _("Sleeping %u second(s)..."), uc_project_get_check_wait ());

  uc_application_display_state_message (CHECK_MESSAGE_LABEL_THIRD, buffer);

  uc_check_wait_value = TRUE;
  id = gtk_timeout_add (uc_project_get_check_wait () * 1000,
			&uc_check_wait_callback, NULL);

  /* sleeping a while... */
  while (uc_check_wait_value)
    {
      if (UC_CHECK_WAIT_END_OTHER_CONDITIONS)
	uc_check_wait_value = FALSE;
      uc_application_update_ui ();
    }

  gtk_timeout_remove (id);
}


/*
 * init a curentItem structure
 */
void
uc_check_currentitem_init (gchar * url, gchar * label,
			   const xmlNodePtr * node)
{
  currentItem.url = url;
  currentItem.label = label;
  currentItem.xml_node = (node) ? *node : NULL;
}


/*
 * make and return a basic UCLinkProperties node
 */
UCLinkProperties *
uc_check_link_properties_node_new (void)
{
  UCLinkProperties *prop = NULL;

  prop = g_new0 (UCLinkProperties, 1);
  prop->uid = uc_check_node_uid++;
  prop->treeview_path = NULL;
  prop->parent = currentItem.parent;
  prop->link_type = LINK_TYPE_NONE;
  prop->link_icon = NULL;
  prop->status_icon = NULL;
  prop->link_value = NULL;
  prop->url = NULL;
  prop->label = NULL;
  prop->h_name = NULL;
  prop->port = NULL;
  prop->path = NULL;
  prop->file = NULL;
  prop->args = NULL;
  prop->domain = NULL;
  prop->header = NULL;
  prop->header_size = 0;
  prop->depth_level = 0;
  prop->metas = NULL;
  prop->emails = NULL;
  prop->all_links = NULL;
  prop->childs = NULL;
  prop->similar_links_parents = NULL;
  prop->tags = NULL;
  prop->is_parsable = FALSE;
  prop->is_downloadable = FALSE;
  prop->xml_node = NULL;
  prop->to_delete = FALSE;
  prop->checked = FALSE;

  return (gpointer) prop;
}


/*
 * process a MAILTO link
 */
static gboolean
uc_check_link_get_properties_proto_mailto (UCLinkProperties * prop)
{
  gchar *email = NULL;
  gchar *tmp = NULL;
  gchar *user = NULL;
  gchar *domain = NULL;

  /* retrieve the host name of the email link */
  uc_utils_split_email (prop->file, &user, &domain);
  prop->domain = domain;
  g_free (user);

  prop->link_type = LINK_TYPE_EMAIL;
  prop->url = g_strdup (prop->file);
  prop->header = g_hash_table_new (g_str_hash, g_str_equal);

  email = strchr (prop->url, ':');
  if (email)
    email++;

  /* check the email address */
  switch (uc_utils_email_is_valid
	  (email, (gboolean) uc_project_get_proto_mailto_check_mx ()))
    {
    case UC_EMAIL_SYNTAX_BAD:
      tmp = STATUS_CODE_BAD_EMAIL_LINK;
      break;
    case UC_EMAIL_SYNTAX_MX_BAD:
      tmp = STATUS_CODE_BAD_EMAIL_LINK_MX;
      break;

    case UC_EMAIL_OK:
      tmp = STATUS_CODE_GOOD_EMAIL_LINK;
      break;

      /* syntax and mx are bad */
    default:
      tmp = STATUS_CODE_BAD_EMAIL_LINK;
    }

  g_hash_table_insert (prop->header, g_strdup ("Status"), g_strdup (tmp));

  return FALSE;

}


/*
 * process a FILE link
 */
static gboolean
uc_check_link_get_properties_proto_file (UCLinkProperties * prop,
					 const gchar * url,
					 const UCLinkType type)
{
  gchar *file_path = NULL;
  GHashTable *tmp = NULL;
  gchar *str = NULL;

  tmp = g_hash_table_new (g_str_hash, g_str_equal);

  if (uc_project_get_proto_file_is_error ())
    g_hash_table_insert (tmp, g_strdup ("Status"),
			 g_strdup (STATUS_CODE_FILE_PROTO_ERROR));
  else
    {
      file_path = g_strconcat (prop->path, prop->file, NULL);

      if (g_file_test (file_path, G_FILE_TEST_EXISTS))
	{
	  str = uc_utils_get_file_content (file_path);
	  uc_cache_append_source (prop->uid, str, strlen (str));
	  g_hash_table_insert (tmp, g_strdup (CONTENT_LENGTH),
			       g_strdup_printf ("%u", strlen (str)));
	  g_hash_table_insert (tmp,
			       g_strdup ("Status"),
			       g_strdup (STATUS_CODE_PAGE_OK));
	}
      else
	g_hash_table_insert (tmp,
			     g_strdup ("Status"),
			     g_strdup (STATUS_CODE_PAGE_NOT_FOUND));
      g_free (file_path);
    }
  g_hash_table_insert (tmp, g_strdup (CONTENT_TYPE),
		       g_strdup (CONTENT_TYPE_HTML));

  prop->header = tmp;

  prop->domain = g_strdup ("");
  prop->url = g_strdup (url);
  prop->link_type = type;

  return FALSE;
}


/*
 * process a HTTP link
 */
static UCHTTPCheckReturn
uc_check_link_get_properties_proto_http (UCLinkProperties * prop,
					 const gchar * url,
					 const gchar * current_path,
					 gchar * current_host,
					 UCHTMLTag * tag)
{
  gchar *str = NULL;
  gchar *status_code = NULL;
  GList *item = NULL;
  gchar *content_type = NULL;
  gchar *normalized_url = NULL;
  gchar *href_base = NULL;
  gboolean already_checked = FALSE;

  /* does a specific port have been
   * specified */
  if (!strcmp (prop->port, "80"))
    prop->url =
      g_strconcat (g_strdown (prop->protocol),
		   "://", prop->h_name,
		   (prop->path[0] != '/' && prop->file[0] != '/') ? "/" : "",
		   prop->path, prop->file, prop->args, NULL);
  else
    prop->url =
      g_strconcat (g_strdown (prop->protocol),
		   "://", prop->h_name, ":", prop->port,
		   (prop->path[0] != '/' && prop->file[0] != '/') ? "/" : "",
		   prop->path, prop->file, prop->args, NULL);

  /* cancel if bad header */
  if (!(prop->header = uc_check_url_get_header ((gpointer) prop)))
    return UC_HTTP_CHECK_RETURN_BAD;

  status_code = g_hash_table_lookup (prop->header, "Status");
  content_type = g_hash_table_lookup (prop->header, CONTENT_TYPE);

  /* it is necessary to correct the link properties
   * after retrived its header, to know exactly
   * if it is a directory or a file */
  uc_url_correction (prop);

  /* cancel if content type not accepted */
  if (!uc_check_content_type_accepted (content_type))
    return UC_HTTP_CHECK_RETURN_BAD;

  prop->link_type = tag->type;
  prop->domain = g_strdup ("");

  /* we want to parse uniquement "text/html"
   * content type */
  if (uc_check_status_is_good (status_code))
    {
      prop->is_parsable = uc_check_content_type_parsable (content_type);

      prop->is_downloadable =
	(prop->is_parsable ||
	 uc_check_content_type_downloadable (content_type));

      if (uc_project_get_url ())
	prop->is_parsable = (prop->is_parsable &&
			     !g_strcasecmp (prop->h_name,
					    uc_project_get_current_host ()));
    }

  /* we don't want to download this link */
  if (!prop->is_downloadable)
    return UC_HTTP_CHECK_RETURN_OK;

  /* we don't want to processed already checked
   * links */
  normalized_url = uc_url_normalize (current_host, current_path, tag->value);
  already_checked =
    uc_check_link_already_checked_with_insert (prop, normalized_url);
  g_free (normalized_url);

  if (already_checked)
    return UC_HTTP_CHECK_RETURN_BAD;

  if (uc_url_is_faked (prop, tag))
    return UC_HTTP_CHECK_RETURN_FAKED_URL;

  /* does this item is authorized? */
  if (!uc_check_apply_filters (prop))
    {
      uc_utils_debug ("        FILTERED: %s\n", prop->url);
      return UC_HTTP_CHECK_RETURN_BAD;
    }

  /* if we have not retrieved the source, something wrong
   * occured or the user hace canceled the check */
  if (!(str = uc_check_url_get_content (prop, current_path, prop->file,
					prop->port, prop->args,
					current_host)))
    return UC_HTTP_CHECK_RETURN_BAD;

  /* check if we have enought spece on the disk to
   * retrieve the source */
  if (!uc_utils_disk_free_enough (UC_CHECK_MINIMUM_SPACE_REQUIRED))
    {
      uc_check_cancel_set_value (TRUE);
      uc_application_dialog_show (_("gURLChecker has canceled the "
				    "check because you does not have "
				    "enough space on your working "
				    "directory partition.\n\n"
				    "gURLChecker has taken a security "
				    "margin to estimate free space."),
				  GTK_MESSAGE_WARNING);
    }

  /* if this is a picture, rewrite it after removing 
   * HTTP header*/
  if (prop->link_type == LINK_TYPE_IMAGE)
    uc_cache_append_source (prop->uid, str,
			    atoi (g_hash_table_lookup
				  (prop->header, CONTENT_LENGTH)));

  /* if the link passed the filters, build node lists
   * for link properties */
  else
    {
      prop->tags = uc_html_parser_get_tags (str, &href_base);

      if (uc_check_ignore_item_get_value ())
	return UC_HTTP_CHECK_RETURN_IGNORE;

      item = prop->tags;
      while (item)
	{
	  UCHTMLTag *tag = (UCHTMLTag *) item->data;

	  uc_application_update_ui ();

	  item = g_list_next (item);

	  if (tag->type == LINK_TYPE_META)
	    prop->metas = g_list_append (prop->metas, (gpointer) tag);
	  else if (tag->type == LINK_TYPE_EMAIL)
	    prop->emails = g_list_append (prop->emails, (gpointer) tag);
	  else if (tag->type != LINK_TYPE_NONE)
	    prop->all_links = g_list_append (prop->all_links, (gpointer) tag);
	}
    }

  g_free (href_base);
  g_free (str);

  return UC_HTTP_CHECK_RETURN_OK;
}


/*
 * fill the UCLinkProperties struct to associate with
 * the url
 */
UCLinkProperties *
uc_check_link_get_properties (const guint depth, gchar * current_host,
			      const gchar * current_path,
			      UCHTMLTag * tag, gboolean * accept)
{
  static guint retry = 0;
  UCLinkProperties *prop = NULL;
  gchar *status = NULL;
  UCStatusCode *sc = NULL;
  gchar *protocol = NULL;
  UCError error = UC_ERROR_NONE;
  gboolean cancel = FALSE;

  /* do we accept to check this protocol? */
  protocol = uc_url_get_protocol (tag->value);
  if (!uc_check_protocol_accepted (protocol))
    {
      g_free (protocol);
      return NULL;
    }

  uc_application_display_state_message (CHECK_MESSAGE_LABEL_SECOND,
					tag->value);
  uc_application_display_state_message (CHECK_MESSAGE_LABEL_THIRD,
					_("Searching..."));
  uc_application_update_ui ();

  prop = uc_check_link_properties_node_new ();
  prop->depth_level = depth;
  prop->parent = currentItem.parent;
  prop->link_value = g_strdup (tag->value);
  prop->protocol = protocol;

  uc_url_parse (current_host, current_path, tag->value, &prop->h_name,
		&prop->port, &prop->path, &prop->file, &prop->args, &error);

  /* href value is empty */
  if (!strcmp (tag->value, " "))
    error = UC_ERROR_MALFORMED_LINK;

  /* if we chroot the check, initialize the directory to
   * chroot in */
  if (!depth)
    uc_project_set_chroot_path ((uc_project_get_check_chroot ())? prop->
				path : NULL);

  /* if we are in chroot mode (just this directory and its subdirectories,
   * or just local links) */
  if (!uc_check_chroot_authorized (prop->h_name, prop->path))
    return NULL;

  /* MAILTO protocol */
  if (tag->type == LINK_TYPE_EMAIL)
    {
      uc_application_display_state_message (CHECK_MESSAGE_LABEL_THIRD,
					    _("Checking email..."));
      uc_application_update_ui ();

      cancel = uc_check_link_get_properties_proto_mailto (prop);
    }

  /* FILE protocol */
  else if (tag->type == LINK_TYPE_FILE_HREF)
    {
      uc_application_display_state_message (CHECK_MESSAGE_LABEL_THIRD,
					    _("Checking local file..."));
      uc_application_update_ui ();

      cancel =
	uc_check_link_get_properties_proto_file (prop, tag->value, tag->type);
    }

  /* HTTP protocol */
  else if (tag->type != LINK_TYPE_NONE)
    {
      UCHTTPCheckReturn ret = UC_HTTP_CHECK_RETURN_OK;

      ret =
	uc_check_link_get_properties_proto_http (prop, tag->value,
						 current_path, current_host,
						 tag);
      cancel = (ret == UC_HTTP_CHECK_RETURN_BAD);

      /* check it is a facked link */
      if (ret == UC_HTTP_CHECK_RETURN_FAKED_URL)
	{
	  if (retry++ <= 2)
	    {
	      uc_lists_checked_links_node_free ((gpointer) prop);

	      prop = uc_check_link_get_properties (depth, current_host,
						   current_path, tag, accept);

	      cancel = FALSE;
	    }
	  else
	    cancel = TRUE;
	}

      retry = 0;
    }
  else
    return NULL;

  /* if a error occured of if the user have
   * cancel the check */
  if (!prop || cancel)
    return NULL;

  /* if user choose to ignore this links */
  if (uc_check_ignore_item_get_value ())
    {
      uc_lists_header_free (prop->header);
      prop->header = uc_check_url_get_header_ignored ();
    }

  *accept = uc_check_apply_filters (prop);

  /* if the link is malformed */
  if (error == UC_ERROR_MALFORMED_LINK)
    {
      tag->malformed = TRUE;
      g_hash_table_replace (prop->header, g_strdup ("Status"),
			    g_strdup (STATUS_CODE_MALFORMED_LINK));
    }

  prop->link_icon = uc_check_get_link_type_icon (prop->link_type);

  if (prop->header)
    {
      status = g_hash_table_lookup (prop->header, "Status");

      ((sc = uc_application_get_status_code_properties (status))
       || (sc =
	   uc_application_get_status_code_properties (STATUS_CODE_UNKNOWN)));
    }

  prop->status_icon = sc->icon_file;
  prop->label = g_strdup (tag->label);

  return prop;
}
