/*
 * Copyright (C) 2002, 2003 Emmanuel Saracco <esaracco@noos.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "html_parser.h"


#define UC_HTML_PARSER_WHILE_CONDITION \
        ( \
                  (!uc_check_cancel_get_value ()) && \
                  (!uc_check_ignore_item_get_value ()) \
        )


static GHashTable *uc_html_parser_get_attributes (gchar * tag);
static gchar *uc_html_parser_format_attribute_name (const gchar * attr);
static gchar *uc_html_parser_format_attribute_value (gchar * attr,
						     const gboolean is_label);
static gboolean uc_html_parser_displayed_value (const gchar * name);
static void uc_html_parser_set_key_value (GHashTable ** attrs, gchar * key,
					  gchar * value);
static gboolean uc_html_parser_authorized_tag_name (const gchar * name);
static gchar *uc_html_parser_prepare_buffer (gchar * str);
static gchar *uc_html_parser_add_base (const gchar * base, const gchar * url);


/*
 * add a key/value pair in the attributes
 * list
 */
static void
uc_html_parser_set_key_value (GHashTable ** attrs, gchar * key, gchar * value)
{
  gchar *clean_key = NULL;
  gchar *clean_value = NULL;

  clean_key = uc_html_parser_format_attribute_name (key);
  clean_value = (uc_html_parser_displayed_value (clean_key)) ?
    uc_html_parser_format_attribute_value (value, TRUE) :
    uc_html_parser_format_attribute_value (value, FALSE);
  g_hash_table_insert (*attrs, clean_key, clean_value);
}


/*
 * return TRUE is the corresponding value of the
 * given attribute will be displayed
 */
static gboolean
uc_html_parser_displayed_value (const gchar * name)
{
  return (name && (!strcmp (name, "ALT") || !strcmp (name, "CONTENT")));
}


/*
 * return TRUE if the given name is usefull
 * for us
 */
static gboolean
uc_html_parser_authorized_tag_name (const gchar * name)
{
  return (name && (!strcmp (name, "IMG") ||
		   !strcmp (name, "A") ||
		   !strcmp (name, "BASE") ||
		   !strcmp (name, "META") ||
		   !strcmp (name, "FRAME") || !strcmp (name, "IFRAME")));
}


/*
 * format a given html attribute value
 */
static gchar *
uc_html_parser_format_attribute_value (gchar * value, const gboolean is_label)
{
  gchar *tmp = NULL;

  if (!value)
    return g_strdup ("");

  if (is_label)
    tmp = uc_utils_format_list_label (value);
  else
    tmp = uc_utils_replacelr (g_strdup (value), ' ');

  return tmp;
}


/*
 * format a given html attribute name
 */
static gchar *
uc_html_parser_format_attribute_name (const gchar * name)
{
  return (!name) ? g_strdup ("") : g_utf8_strup (name, -1);
}


/*
 * prepare the buffer before parsing it, to avoid
 * traps and other bad things...
 */
static char *
uc_html_parser_prepare_buffer (gchar * str)
{
  GString *tmp = g_string_new ("");
  gchar *ret = NULL;
  guint check_comments = 0;

  /* if user want to check HTML comments, remove
   * their tags from the source buffer */
  check_comments = uc_project_get_check_comments ();

  /* skip HTTP header */
  str = uc_utils_search_string_next (str, "\r\n\r\n", '<');

  /* remove new line before a > */
  while (*str && (str + 4))
    {
      if ((check_comments) && !memcmp (&(*str), "<!--", 4))
	tmp = g_string_append_c (tmp, ' '), str += 3;
      else if ((check_comments) && !memcmp (&(*str), "-->", 3))
	tmp = g_string_append_c (tmp, ' '), str += 2;
      else if (!
	       ((*str == '\n' || *str == '\r') &&
		(g_ascii_ispunct (*(str - 1)) ||
		 g_ascii_isspace (*(str - 1)))))
	tmp = g_string_append_c (tmp, *str);

      str++;
    }

  ret = tmp->str;
  g_string_free (tmp, FALSE);

  return ret;
}


/*
 * parse the given buffer to extract HTML tags and
 * return a glist
 */
GList *
uc_html_parser_get_tags (gchar * buffer_orig, gchar ** href_base)
{
  gchar *buffer = NULL;
  gchar *buffer_begin = NULL;
  gchar *begin = NULL;
  gchar *end = NULL;
  gchar *label_begin = NULL;
  GList *tags = NULL;
  GList *item = NULL;
  UCHTMLTag *previous = NULL;
  gchar *previous_label = NULL;
  gchar *tmp = NULL;
  gchar label[BUFFER_URL_LEN + 1];
  guint i = 0;
  guint nbr_tags = 0;

  uc_application_display_state_message (CHECK_MESSAGE_LABEL_THIRD,
					_("Parsing HTML code..."));
  uc_application_update_ui ();

  buffer = uc_html_parser_prepare_buffer (buffer_orig);
  buffer_begin = buffer;

  do
    {
      g_free (tmp);
      tmp = NULL;

      if (!*buffer ||
	  !(begin = strpbrk (buffer, "<")) ||
	  !*(++begin) ||
	  !(end = strpbrk (begin, ">")) ||
	  !(tmp = uc_utils_strdup_delim (begin, end)))
	break;

      /* pre selection of what we want to proceed */
      if (!strchr ("!SsIiAaBbMmFf", tmp[0]))
	{
	  buffer++;
	  continue;
	}

      /* parse comments? */
      if (!uc_project_get_check_comments ()
	  && uc_utils_memcasecmp (tmp, "!--"))
	{
	  buffer = begin;

	  while (*buffer && (!uc_utils_memcasecmp (buffer, "-->")))
	    buffer++;
	}

      /* we don't want any scripting element */
      else if (uc_utils_memcasecmp (tmp, "SCRIPT"))
	{
	  buffer = begin;

	  while (*buffer && (!uc_utils_memcasecmp (buffer, "/SCRIPT>")))
	    buffer++;
	}
      else if (uc_utils_memcasecmp (tmp, "STYLE"))
	{
	  buffer = begin;

	  while (*buffer && (!uc_utils_memcasecmp (buffer, "/STYLE>")))
	    buffer++;
	}
      else
	{
	  label_begin = end + 1;
	  i = 0;

	  if (uc_utils_memcasecmp (tmp, "A "))
	    while (*label_begin
		   &&
		   (!uc_utils_memcasecmp
		    (label_begin, "</A>")) && (i < BUFFER_URL_LEN))
	      label[i++] = *label_begin++;
	  else
	    while (*label_begin
		   && (*label_begin != '<') && (i < BUFFER_URL_LEN))
	      label[i++] = *label_begin++;

	  /* we don't want end tags */
	  if (*tmp && (*tmp != '/'))
	    {
	      UCHTMLTag *tb = NULL;
	      gchar *tmp1 = NULL;
	      gchar *tmp2 = NULL;

	      /* previous item label */
	      if (previous && previous_label && strlen (previous_label))
		{
		  g_free (previous->label);
		  previous->label =
		    uc_utils_format_list_label (previous_label);
		}
	      g_free (previous_label);

	      if (strchr (tmp, ' '))
		{
		  tmp2 = uc_utils_strdup_delim (tmp, strpbrk (tmp, " "));
		  tmp1 = uc_utils_replacelr (g_utf8_strup (tmp2, -1), ' ');
		  g_free (tmp2);
		}
	      else
		tmp1 = uc_utils_replacelr (g_utf8_strup (tmp, -1), ' ');

	      /* register this tag in our hash only if it is a tag
	       * we want to treat (typically: IMG, A, META, FRAME or IFRAME) */
	      if (uc_html_parser_authorized_tag_name (tmp1))
		{
		  gchar msg[255 + 1] = { 0 };

		  tb = uc_html_parser_node_new (NULL);

		  nbr_tags++;

		  snprintf (msg, 255,
			    _("Parsing HTML code "
			      "(founded <b>%u</b> links)..."), nbr_tags);
		  uc_application_display_state_message
		    (CHECK_MESSAGE_LABEL_THIRD, msg);
		  uc_application_update_ui ();

		  tb->name = g_strdup (tmp1);
		  tb->label = g_strdup ("");
		  tb->attr = uc_html_parser_get_attributes (tmp);
		  tags = g_list_append (tags, (gpointer) tb);
		  previous = (gpointer) tb;
		}

	      g_free (tmp1);

	      label[i] = '\0';
	      previous_label = g_strdup (label);
	    }

	  if (*buffer && !strstr (label, "<img ") && !strstr (label, "<IMG "))
	    buffer += strlen (label);
	  if (*buffer)
	    buffer += (end - begin) + 2;
	}
    }
  while (UC_HTML_PARSER_WHILE_CONDITION);

  /* previous item label */
  if (previous && previous_label && strlen (previous_label))
    {
      g_free (previous->label);
      previous->label = uc_utils_format_list_label (previous_label);
    }
  g_free (previous_label);

  g_free (buffer_begin);

  item = g_list_first (tags);
  while (item)
    {
      UCHTMLTag *tag = NULL;
      gchar *str_label = NULL;
      gchar *str_value = NULL;

      tag = (UCHTMLTag *) item->data;
      item = g_list_next (item);

      if (!tag)
	{
	  g_warning ("tag is null!");
	  continue;
	}

      if (!*href_base && !strcmp (tag->name, "BASE")
	  && (*href_base =
	      g_strdup (g_hash_table_lookup (tag->attr, "HREF"))))
	{
	  if (strlen (*href_base) >= 2
	      && (*href_base)[strlen (*href_base) - 1] == '/'
	      && (*href_base)[strlen (*href_base) - 2] == '/')
	    (*href_base)[strlen (*href_base) - 1] = '\0';
	}
      else if (!strcmp (tag->name, "META"))
	{
	  str_value = g_hash_table_lookup (tag->attr, "CONTENT");
	  ((str_label = g_hash_table_lookup (tag->attr, "NAME")) ||
	   (str_label = g_hash_table_lookup (tag->attr, "HTTP-EQUIV")));

	  tag->type = LINK_TYPE_META;
	  g_free (tag->label);
	  tag->label = (str_label) ? g_strdup (str_label) : g_strdup ("");
	  tag->value =
	    (str_value) ? g_convert (str_value, -1, "UTF-8",
				     uc_project_get_local_charset (),
				     NULL, NULL, NULL) : g_strdup ("");
	}
      else if (!strcmp (tag->name, "IMG"))
	{
	  str_value = g_hash_table_lookup (tag->attr, "SRC");
	  ((str_label = g_hash_table_lookup (tag->attr, "ALT")) ||
	   (str_label = g_hash_table_lookup (tag->attr, "TITLE")));

	  tag->type = LINK_TYPE_IMAGE;
	  g_free (tag->label);
	  tag->label = (str_label) ? g_strdup (str_label) : g_strdup ("");

	  if (str_value)
	    {
	      tmp = uc_html_parser_add_base (*href_base, str_value);
	      tag->value = g_convert (tmp, -1, "UTF-8",
				      uc_project_get_local_charset (),
				      NULL, NULL, NULL);
	      g_free (tmp);
	    }
	  else
	    tag->value = g_strdup ("");

	  tag->malformed = (!str_value);
	}
      else if (!strcmp (tag->name, "FRAME") || !strcmp (tag->name, "IFRAME"))
	{
	  str_value = g_hash_table_lookup (tag->attr, "SRC");

	  tag->type = LINK_TYPE_FRAME;

	  if (str_value)
	    tag->value =
	      (strlen (str_value)) ?
	      g_strstrip (g_convert
			  (str_value, -1, "UTF-8",
			   uc_project_get_local_charset (),
			   NULL, NULL, NULL)) : g_strdup ("");
	  else
	    tag->value = g_strdup ("");

	  tag->malformed = (!strlen (tag->value));
	}
      else if (!strcmp (tag->name, "A")
	       && (str_value = g_hash_table_lookup (tag->attr, "HREF")))
	{
	  if (uc_utils_memcasecmp (str_value, "MAILTO:"))
	    {
	      tag->type = LINK_TYPE_EMAIL;
	      tag->value =
		g_convert (str_value, -1, "UTF-8",
			   uc_project_get_local_charset (), NULL, NULL, NULL);
	    }
	  else if (uc_utils_memcasecmp (str_value, "FILE:"))
	    {
	      tag->type = LINK_TYPE_FILE_HREF;
	      tag->value =
		g_convert (str_value, -1, "UTF-8",
			   uc_project_get_local_charset (), NULL, NULL, NULL);
	    }
	  else
	    {
	      gchar *tmp = NULL;

	      tag->type = LINK_TYPE_HREF;

	      tmp = uc_html_parser_add_base (*href_base, str_value);
	      tag->value =
		g_convert (tmp, -1, "UTF-8", uc_project_get_local_charset (),
			   NULL, NULL, NULL);
	      g_free (tmp);
	    }
	}
    }

  return tags;
}


/*
 * add the base tag value to the link if necessary
 */
static gchar *
uc_html_parser_add_base (const gchar * base, const gchar * url)
{
  gchar *path = NULL;
  gchar *file = NULL;
  gchar *new = NULL;
  gboolean is_file = FALSE;
  gboolean is_args = FALSE;

  path = g_strdup (base);

  if (!path || !url || strstr (url, "://") || !(file = strrchr (path, '/')))
    {
      g_free (path);
      return g_strdup (url);
    }

  if (*(file - 1) == '/')
    file = NULL;
  else
    {
      *file = '\0';
      file++;
    }

  is_file = (path[strlen (path) - 1] != '/');
  is_args = (strchr ("?#", url[0]) != NULL);

  if ((path[strlen (path) - 1]) == '/')
    path[strlen (path) - 1] = '\0';

  new = (!is_args) ?
    g_strconcat (path, ((url[0] != '/') ? "/" : ""), url, NULL) :
    g_strconcat (path, "/", file, url, NULL);

  g_free (path);

  return new;
}


/*
 * create a new tag list node
 */
UCHTMLTag *
uc_html_parser_node_new (gpointer name)
{
  UCHTMLTag *item = NULL;

  item = g_new0 (UCHTMLTag, 1);
  item->depth = 0;
  item->name = (gchar *) name;
  item->label = NULL;
  item->value = NULL;
  item->attr = NULL;
  item->type = LINK_TYPE_NONE;
  item->malformed = FALSE;

  return item;
}


/*
 * return attributes of a tag in a 
 * g_hash_table
 */
static GHashTable *
uc_html_parser_get_attributes (gchar * tag)
{
  GHashTable *attrs = NULL;
  guint i = 0;
  gchar key[BUFFER_URL_LEN + 1] = { 0 };
  gchar value[BUFFER_URL_LEN + 1] = { 0 };
  gboolean simple_quote = FALSE;
  gboolean double_quote = FALSE;
  gboolean is_value = FALSE;
  gboolean new_item = FALSE;

  attrs = g_hash_table_new (g_str_hash, g_str_equal);
  if (!tag)
    return attrs;

  tag = uc_utils_replacelr (tag, ' ');
  if (strchr (tag, ' '))
    {
      UCHTMLTag *html_tag = NULL;
      gchar *end = NULL;

      end = strpbrk (tag, " ");
      html_tag = uc_html_parser_node_new (uc_utils_strdup_delim (tag, end));

      tag += (end - tag) + 1;
      while (*tag && (i < BUFFER_URL_LEN))
	{
	  switch (*tag)
	    {
	      /* double quote */
	    case '"':

	      if (double_quote)
		new_item = TRUE, double_quote = FALSE;
	      else if (simple_quote)
		value[i++] = *tag;
	      else
		{
		  if (*(tag + 1) && (*(tag + 1) == '"'))
		    {
		      new_item = TRUE;
		      double_quote = FALSE;
		      value[i++] = ' ';
		      tag++;
		    }
		  else
		    double_quote = TRUE;
		}
	      break;

	      /* simple quote */
	    case '\'':

	      if (simple_quote)
		new_item = TRUE, simple_quote = FALSE;
	      else if (double_quote)
		value[i++] = *tag;
	      else
		{
		  if (*(tag + 1) && (*(tag + 1) == '\''))
		    {
		      new_item = TRUE;
		      simple_quote = FALSE;
		      value[i++] = ' ';
		      tag++;
		    }
		  else
		    simple_quote = TRUE;
		}
	      break;

	      /* space or tabulation */
	    case '\t':
	    case ' ':

	      if ((!double_quote) && (!simple_quote))
		new_item = TRUE;
	      else
		{
		  if (is_value)
		    value[i] = *tag;
		  else
		    key[i] = *tag;
		  i++;
		}
	      break;

	      /* equal */
	    case '=':

	      if ((!double_quote) && (!simple_quote) && (!is_value))
		new_item = is_value = TRUE;
	      else
		value[i++] = *tag;
	      break;

	      /* other */
	    default:

	      if (is_value)
		value[i] = *tag;
	      else
		key[i] = *tag;
	      i++;

	    }

	  tag++;

	  if (new_item)
	    {
	      if (strlen (value))
		{
		  uc_html_parser_set_key_value (&attrs, key, value);
		  memset (&key, 0, BUFFER_URL_LEN + 1);
		  is_value = FALSE;
		}

	      i = 0;
	      new_item = FALSE;
	      memset (&value, 0, BUFFER_URL_LEN + 1);

	      double_quote = FALSE;
	      simple_quote = FALSE;
	    }
	}

      if (strlen (value))
	uc_html_parser_set_key_value (&attrs, key, value);
    }

  return attrs;
}
