/* 
 * Copyright (C) 2002, 2003 Emmanuel Saracco <esaracco@noos.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA. 
 */

#include "lists.h"


/*
 * struct used for the uc_lists_already_checked_links_list
 * list
 */
typedef struct
{
  gpointer url;
  UCLinkProperties *parent;
}
UCAlreadyChecked;


/* temporary list of already checked urls */
static GList *uc_lists_already_checked_links_list = NULL;

/* temporary list of already checked urls */
static GList *uc_lists_checked_links_list = NULL;

static void uc_lists_checked_links_free_real (GList * list);
static void uc_lists_checked_links_free_callback (gpointer key,
						  gpointer value,
						  gpointer user_data);
static UCAlreadyChecked *uc_lists_already_checked_links_node_new (const
								  gchar *
								  url,
								  gpointer
								  ac_p);
static void uc_lists_already_checked_free_callback (gpointer data,
						    gpointer user_data);
static UCLinkProperties *uc_lists_checked_links_lookup_by_uid_real (GList *
								    list,
								    const
								    guint
								    uid);

/*
 * free a html tags list
 */
void
uc_lists_html_tags_free (GList * list)
{
  GList *item = NULL;
  GHashNode *node = NULL;
  guint i = 0;

  if (!list)
    return;

  item = list;
  while (item)
    {
      UCHTMLTag *tag = (UCHTMLTag *) item->data;

      g_free (tag->name);
      g_free (tag->label);
      g_free (tag->value);

      if (tag->attr)
	{
	  for (i = 0; i < tag->attr->size; i++)
	    for (node = tag->attr->nodes[i]; node; node = node->next)
	      g_free (node->key), g_free (node->value);

	  g_hash_table_destroy (tag->attr);
	}

      item = g_list_next (item);
    }

  g_list_free (list);
  list = NULL;
}


/*
 * free the given similar links list
 */
void
uc_lists_similar_links_free (gpointer list)
{
  GList *item = NULL;

  if (!list)
    return;

  item = (GList *) list;
  while (item)
    {
      g_free ((gchar *) item->data);

      item = g_list_next (item);
    }

  g_list_free ((GList *) list);
  (GList *) list = NULL;
}


/*
 * append a item to the uc_lists_already_checked_links_list
 * list
 */
void
uc_lists_already_checked_links_append (const gchar * url, gpointer ac_p)
{
  uc_lists_already_checked_links_list =
    g_list_prepend (uc_lists_already_checked_links_list,
		    uc_lists_already_checked_links_node_new (url,
							     (gpointer)
							     ac_p));
}


/*
 * return TRUE if the list of checked
 * links is empty
 */
guint
uc_lists_checked_links_is_empty (void)
{
  return (!g_list_length (uc_lists_checked_links_list));
}


/*
 * public wrapper to freeing the uc_lists_checked_links_list
 * list
 */
void
uc_lists_checked_links_free ()
{
  uc_lists_checked_links_free_real (uc_lists_checked_links_list);
}


/*
 * delete a page header
 */
void
uc_lists_header_free (GHashTable * header)
{
  g_hash_table_foreach (header, uc_lists_checked_links_free_callback, NULL);
  g_hash_table_destroy (header);
}


/*
 * free a UCLinkProperties node
 */
void
uc_lists_checked_links_node_free (UCLinkProperties * prop)
{
  if (!prop)
    return;

  g_free (prop->treeview_path);
  if (prop->link_icon)
    g_object_unref (prop->link_icon);
  g_free (prop->link_value);
  g_free (prop->label);
  g_free (prop->url);
  g_free (prop->protocol);
  g_free (prop->h_name);
  g_free (prop->port);
  g_free (prop->path);
  g_free (prop->domain);
  g_free (prop->file);
  g_free (prop->args);

  if (prop->header)
    uc_lists_header_free (prop->header);

  if (prop->childs)
    uc_lists_checked_links_free_real (prop->childs);

  if (prop->similar_links_parents)
    uc_lists_similar_links_free (prop->similar_links_parents);

  /* prop->all_links (ptr on prop->tags items) */
  /* prop->emails (ptr on prop->tags items) */
  /* prop->metas ((ptr on prop->tags items) */

  if (prop->tags)
    uc_lists_html_tags_free (prop->tags);

  g_free (prop);
}


/*
 * free all the allocated objects of the
 * url properties list
 */
static void
uc_lists_checked_links_free_real (GList * list)
{
  GList *item = NULL;

  if (!list)
    return;

  item = list;
  while (item)
    {
      uc_lists_checked_links_node_free ((UCLinkProperties *) item->data);

      item = g_list_next (item);
    }

  g_list_free (list);
  list = NULL;
}


/*
 * TRUE if the given url already in
 * the checked urls list
 */
gboolean
uc_lists_checked_links_exist (gchar * url)
{
  return (uc_lists_checked_links_lookup (url) != NULL);
}


/*
 * return a pointer on the found item
 */
gpointer
uc_lists_checked_links_lookup (gchar * url)
{
  return (uc_lists_already_checked_links_lookup (url));
}


/*
 * TRUE if the given url already in
 * the checked urls list
 */
gboolean
uc_lists_already_checked_links_exist (gchar * url)
{
  return (uc_lists_already_checked_links_lookup (url) != NULL);
}


/*
 * return a pointer on the found item
 */
gpointer
uc_lists_already_checked_links_lookup (gchar * url)
{
  GList *item = NULL;

  item = uc_lists_already_checked_links_list;
  while (item)
    {
      UCAlreadyChecked *ac = (UCAlreadyChecked *) item->data;

      item = g_list_next (item);
      if (!ac)
	continue;

      if (!strcmp (ac->url, url))
	return (gpointer) ac->parent;
    }

  return NULL;
}


/*
 * append a url in the similar links
 * list
 */
void
uc_lists_similar_links_append (gpointer prop_p, gpointer url_p)
{
  UCLinkProperties *prop = (UCLinkProperties *) prop_p;

  if (!prop)
    return;

  prop->similar_links_parents =
    g_list_prepend (prop->similar_links_parents, g_strdup (url_p));
}


/*
 * callback for uc_check_already_checked_links_free
 */
static void
uc_lists_already_checked_free_callback (gpointer data, gpointer user_data)
{
  UCAlreadyChecked *ac = (UCAlreadyChecked *) data;

  if (ac)
    g_free (ac->url);
}


/*
 * free all allocated url strings
 */
void
uc_lists_already_checked_free (void)
{
  if (!uc_lists_already_checked_links_list)
    return;

  g_list_foreach (uc_lists_already_checked_links_list,
		  uc_lists_already_checked_free_callback, NULL);

  g_list_free (uc_lists_already_checked_links_list);
  uc_lists_already_checked_links_list = NULL;
}


/*
 * funtion passed to g_hash_table_foreach() to
 * free all the allocated object
 */
static void
uc_lists_checked_links_free_callback (gpointer key,
				      gpointer value, gpointer user_data)
{
  g_free (key);
  g_free (value);
}


/*
 * allocate and fill a uc_lists_already_checked_links_list node
 */
static UCAlreadyChecked *
uc_lists_already_checked_links_node_new (const gchar * url, gpointer ac_p)
{
  UCAlreadyChecked *ac = NULL;

  ac = g_new0 (UCAlreadyChecked, 1);

  ac->url = g_strdup (url);
  ac->parent = (gpointer) ac_p;

  return ac;
}


/*
 * return properties of the given checked url
 * uid
 */
UCLinkProperties *
uc_lists_checked_links_lookup_by_uid (const guint uid)
{
  return uc_lists_checked_links_lookup_by_uid_real
    (uc_lists_checked_links_list, uid);
}


/*
 * set the already checked urls list
 */
void
uc_lists_checked_links_set (GList * list)
{
  uc_lists_checked_links_list = list;
}


/*
 * return the already checked urls list
 */
GList *
uc_lists_checked_links_get (void)
{
  return uc_lists_checked_links_list;
}


/*
 * recursively search a item in the
 * url properties list
 */
static UCLinkProperties *
uc_lists_checked_links_lookup_by_uid_real (GList * list, const guint uid)
{
  GList *item = NULL;
  UCLinkProperties *dum = NULL;
  UCLinkProperties *lp = NULL;

  if (!list)
    return NULL;

  item = list;
  while (item)
    {
      lp = (UCLinkProperties *) item->data;

      item = g_list_next (item);
      if (!lp)
	continue;

      if (lp->uid == uid)
	return lp;

      if ((lp->childs) &&
	  ((dum =
	    uc_lists_checked_links_lookup_by_uid_real (lp->childs, uid))))
	return dum;
    }

  return NULL;
}


/*
 * return TRUE if the given url already exist in
 * the given similar links list
 */
gboolean
uc_lists_similar_links_exist (GList * list, const gchar * url)
{
  GList *item = NULL;

  if (!list)
    return FALSE;

  item = list;
  while (item)
    {
      gchar *str = (gchar *) item->data;

      item = g_list_next (item);
      if (!str)
	continue;

      if (!strcmp (str, url))
	return TRUE;
    }

  return FALSE;
}
