/*
 * Copyright (C) 2002, 2003 Emmanuel Saracco <esaracco@noos.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "report.h"


/* structure to store informations during the
   check for managing a project report */
static struct UCProjectReport
{
  guint allLinks;
  guint checkedLinks;
  guint badLinks;
  guint goodLinks;
  guint timedOutLinks;
  guint ignoredLinks;
  guint malformedLinks;
  gdouble elapsedTime;
}
projectReport;


static void report_set_progress_bar (const guint report_item,
				     const gchar * glade_label,
				     const gchar * report_label,
				     const gchar * color);


/*
 * return the total of links
 */
guint
uc_report_get_alllinks (void)
{
  return projectReport.allLinks;
}

/*
 * return the number of checked
 * links
 */
guint
uc_report_get_checkedlinks (void)
{
  return projectReport.checkedLinks;
}


/*
 * set "malformed links" value
 */
void
uc_report_set_malformedlinks (const guint val)
{
  projectReport.malformedLinks += (val) ? val : 1;
}


/*
 * set "ignored links" value
 */
void
uc_report_set_ignoredlinks (const guint val)
{
  projectReport.ignoredLinks += (val) ? val : 1;
}


/*
 * set "checked links" value
 */
void
uc_report_set_alllinks (const guint val)
{
  projectReport.allLinks += (val) ? val : 1;
}


/*
 * set "good links" value
 */
void
uc_report_set_goodlinks (const guint val)
{
  projectReport.goodLinks += (val) ? val : 1;
}


/*
 * set "bad links" value
 */
void
uc_report_set_badlinks (const guint val)
{
  projectReport.badLinks += (val) ? val : 1;
}


/*
 * set "timeouts" value
 */
void
uc_report_set_timedoutlinks (const guint val)
{
  projectReport.timedOutLinks += (val) ? val : 1;
}


/*
 * set "single links" value
 */
void
uc_report_set_checkedlinks (void)
{
  projectReport.checkedLinks =
    projectReport.timedOutLinks +
    projectReport.badLinks +
    projectReport.goodLinks +
    projectReport.ignoredLinks + projectReport.malformedLinks;
}


/*
 * reset the check report structure
 */
void
uc_check_report_reset (void)
{
  projectReport.allLinks = 0;
  projectReport.checkedLinks = 0;
  projectReport.badLinks = 0;
  projectReport.goodLinks = 0;
  projectReport.ignoredLinks = 0;
  projectReport.timedOutLinks = 0;
  projectReport.malformedLinks = 0;
  projectReport.elapsedTime = 0;

  uc_report_display_update ();
}


/*
 * update the display check report
 */
void
uc_report_display_update (void)
{
  gchar *tmp = NULL;
  GtkWidget *widget = NULL;

  uc_report_set_checkedlinks ();

  /* checked links */
  tmp = (uc_project_get_bookmarks_file ())?
    g_strdup_printf (_("<i><b>Checked links</b></i>: %d"),
		     projectReport.checkedLinks) :
    g_strdup_printf (_("<i><b>Checked links</b></i>: %d / %d"),
		     projectReport.checkedLinks, projectReport.allLinks);

  UC_GLADE_GET_WIDGET ("label_checked_links", glade_main_window, widget);
  gtk_label_set_markup (GTK_LABEL (widget), tmp);
  g_free (tmp);

  /* good links */
  report_set_progress_bar (projectReport.goodLinks,
			   "pb_good_links",
			   _("Good links"), GOOD_LINK_BGCOLOR);

  /* bad links */
  report_set_progress_bar (projectReport.badLinks,
			   "pb_bad_links", _("Bad links"), BAD_LINK_BGCOLOR);

  /* ignored links */
  report_set_progress_bar (projectReport.ignoredLinks,
			   "pb_ignored_links",
			   _("Ignored links"), IGNORED_LINK_BGCOLOR);

  /* malformed links */
  report_set_progress_bar (projectReport.malformedLinks,
			   "pb_malformed_links",
			   _("Malformed links"), MALFORMED_LINK_BGCOLOR);

  /* timeouts */
  report_set_progress_bar (projectReport.timedOutLinks,
			   "pb_timeouts",
			   _("Timeouts"), TIMEOUT_LINK_BGCOLOR);

  /* ellapsed time initialization */
  if (!projectReport.checkedLinks)
    {
      UC_GLADE_GET_WIDGET ("label_ellapsed_time", glade_main_window, widget);
      gtk_label_set_markup (GTK_LABEL (widget),
			    _("<i><b>Elapsed time</b></i>: 00:00:00"));
    }

  uc_application_update_ui ();
}


/*
 * set label and progression on a report
 * progress bar
 */
static void
report_set_progress_bar (const guint report_item, const gchar * glade_label,
			 const gchar * report_label, const gchar * bg_color)
{
  gint percent = 0;
  GtkWidget *widget = NULL;
  gchar *tmp = NULL;
  GdkColor color;

  percent =
    (projectReport.checkedLinks) ? (report_item * 100) /
    projectReport.checkedLinks : 0;

  UC_GLADE_GET_WIDGET (glade_label, glade_main_window, widget);

  /* set progress bar bg color */
  gdk_color_parse (bg_color, &color);
  gtk_widget_modify_bg (widget, GTK_STATE_PRELIGHT, &color);

  /* set progress bar fg color */
  gdk_color_parse ("black", &color);
  gtk_widget_modify_fg (widget, GTK_STATE_PRELIGHT, &color);

  gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (widget),
				 (gfloat) (percent * 0.01));

  tmp = g_strdup_printf ("%s %d (%d%%)", report_label, report_item, percent);
  gtk_progress_bar_set_text (GTK_PROGRESS_BAR (widget), tmp);
  g_free (tmp);

}


/*
 * callback called to display the elapsed
 * time of a check
 */
gboolean
uc_report_timer_callback (gpointer timer)
{
  time_t t;
  struct tm *td;
  GtkWidget *widget = NULL;
  gchar *tmp = NULL;

  if (!uc_check_gdk_input_timeout_locked ())
    uc_check_gdk_input_timeout_increment ();

  time (&t);
  t -= ((time_t) timer + 3600);
  td = localtime (&t);

  UC_GLADE_GET_WIDGET ("label_ellapsed_time", glade_main_window, widget);
  tmp = g_strdup_printf (_
			 ("<i><b>Elapsed time</b></i>: %02d:%02d:%02d"),
			 td->tm_hour, td->tm_min, td->tm_sec);
  gtk_label_set_markup (GTK_LABEL (widget), tmp);
  g_free (tmp);

  return TRUE;
}
