/*
 * Copyright (C) 2002, 2003 Emmanuel Saracco <esaracco@noos.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "url.h"


static gchar *uc_url_get_absolute (gchar * url, UCError * error);


/*
 * TRUE if the link is in fact a redirection or if it
 * has another location
 *
 * -> tag struct is modified
 */
gboolean
uc_url_is_faked (UCLinkProperties * prop, UCHTMLTag * tag)
{
  gchar *new_location = NULL;
  gboolean ret = FALSE;
  gboolean concat = FALSE;

  if (!prop || !prop->header)
    return ret;

  ((new_location = g_hash_table_lookup (prop->header, "Location")) ||
   (new_location = g_hash_table_lookup (prop->header, "location")) ||
   (new_location = g_hash_table_lookup (prop->header, "Content-Location")));

  /* if new_location is ok, go one, else
   * continue to default */
  if (new_location)
    {
      concat =
	(!(uc_utils_memcasecmp (new_location, "HTTP")
	   || tag->value[0] == '/'));

      g_free (tag->value);

      if (new_location[0] == '/')
	tag->value =
	  g_strconcat (g_strdown (prop->protocol), "://",
		       prop->h_name, new_location, NULL);
      else
	{
	  gchar *tmp = NULL;

	  if ((tmp = strrchr (prop->path, '/')))
	    *(tmp + 1) = '\0';

	  if (!concat &&
	      new_location[0] != '/' &&
	      !uc_utils_memcasecmp (new_location, "HTTP"))
	    concat = TRUE;

	  if (concat)
	    tag->value =
	      g_strconcat (g_strdown (prop->protocol), "://", prop->h_name,
			   (prop->path[0] != '/'
			    && prop->file[0] != '/') ? "/" : "", prop->path,
			   ((new_location[0] != '/')
			    && (prop->path[strlen (prop->path) - 1] !=
				'/')) ? "/" : "", new_location, NULL);
	  else if (uc_utils_memcasecmp (new_location, "HTTP"))
	    tag->value = g_strdup (new_location);
	  else
	    tag->value = g_strconcat ("/", new_location, NULL);

	  ret = TRUE;
	}

      uc_utils_debug
	("              Real location: \"%s\" for the following link:\n",
	 tag->value);
    }

  return ret;
}


/*
 * encode a given URL
 *
 * -> function adapted from
 *    http://cvs.php.net/co.php/curl/lib/escape.c
 *   
 */
gchar *
uc_url_encode (const gchar * string)
{
  guint alloc = 0;
  guint newlen = 0;
  guint length = 0;
  guint index = 0;
  gchar in = 0;
  gchar *ns = NULL;
  gchar *testing_ptr = NULL;
  gchar *tmp = NULL;
  gchar *tmp_begin = NULL;

  if (!string || !strlen (string))
    return g_strdup ("");

  /* if a '%' is present, we suppose that the string
   * is already encoded */
  if (strchr (string, '%'))
    return g_strdup (string);

  /* if the string is in UTF-8, decode it before */
  tmp = (g_utf8_validate (string, -1, NULL)) ?
    g_convert (string, -1, uc_project_get_local_charset (), "UTF-8",
	       NULL, NULL, NULL) : g_strdup (string);

  tmp_begin = tmp;

  alloc = newlen = length = strlen (tmp) + 1;
  ns = g_malloc0 (alloc);

  length = alloc - 1;
  while (length--)
    {
      in = *tmp;
      if (!(in >= 'a' && in <= 'z') &&
	  !(in >= 'A' && in <= 'Z') &&
	  !(in >= '0' && in <= '9') && !strchr ("@~:-_./+", in))
	{
	  newlen += 2;

	  if (newlen > alloc)
	    {
	      alloc *= 2;

	      if (!(testing_ptr = g_realloc (ns, alloc)))
		{
		  g_free (ns);
		  return g_strdup ("");
		}
	      else
		ns = testing_ptr;
	    }
	  sprintf (&ns[index], "%%%02X", in);

	  index += 3;
	}
      else
	ns[index++] = in;

      tmp++;
    }

  ns[index] = '\0';

  g_free (tmp_begin);

  /* reencode the string in UTF-8 */
  tmp =
    g_convert (ns, -1, "UTF-8", uc_project_get_local_charset (), NULL,
	       NULL, NULL);
  g_free (ns);

  return tmp;
}


/*
 * return the absolute path. for example:
 *
 * http://www.perl.com/first/second/third/../../.././index.html will become
 * http://www.perl.com/index.html
 */
static gchar *
uc_url_get_absolute (gchar * url, UCError * error)
{
  gchar *ret = NULL;
  gchar *path = NULL;
  gchar *file = NULL;
  gchar *tmp = NULL;
  gchar *url_tmp = NULL;
  gchar *p_last = NULL;
  gchar *p1 = NULL;
  gchar *p2 = NULL;
  gchar *hostname = NULL;
  gchar *protocol = NULL;
  gchar *hostname_tmp = NULL;


  /* malformed URL */
  if (strstr (url, "..."))
    {
      *error = UC_ERROR_MALFORMED_LINK;
      return g_strdup (url);
    }

  if (!(p2 = strstr (url, "./")))
    return g_strdup (url);

  if (!(hostname = uc_url_get_hostname (NULL, url)))
    return g_strdup (url);

  if (!(protocol = uc_url_get_protocol (url)))
    {
      g_free (hostname);
      return g_strdup (url);
    }

  /* malformed URL */
  if (ispunct (hostname[strlen (hostname) - 1]))
    {
      g_free (hostname);
      g_free (protocol);

      *error = UC_ERROR_MALFORMED_LINK;

      return g_strdup (url);
    }

  hostname_tmp = g_strconcat (protocol, "://", hostname, NULL);
  g_free (hostname);
  g_free (protocol);

  if (*(p2 - 1) != '.')
    {
      p1 = url;

      if (*(p2 - 1) == '/')
	{
	  gchar *tmp1 = uc_utils_strdup_delim (p1, p2);
	  gchar *tmp2 = uc_utils_strdup_delim ((p2 + 2),
					       &url[strlen (url)]);

	  url_tmp = g_strconcat (tmp1, tmp2, NULL);

	  g_free (tmp1);
	  g_free (tmp2);
	}
      else
	url_tmp = uc_utils_strdup_delim ((p2 + 2), &url[strlen (url)]);
    }
  else
    url_tmp = g_strdup (url);

  if ((!(p2 = strstr (url_tmp, "../"))) || (*p2 == *url_tmp))
    return url_tmp;

  p1 = url_tmp;
  p2 -= 1;
  path = uc_utils_strdup_delim (p1, p2);
  file = g_strdup (p2);
  p1 = file;

  /* malformed URL */
  if (!strcasecmp (hostname_tmp, path))
    {
      g_free (hostname_tmp);
      g_free (path);
      g_free (file);

      *error = UC_ERROR_MALFORMED_LINK;

      return url_tmp;
    }

  // if the nest is not ok, do not proceed
  if ((tmp = strrchr (path, '/')) && (*(tmp - 1) != '/'))
    {
      while ((p1 = strstr (p1, "../")))
	{
	  p1 += 3;
	  if (strrchr (path, '/'))
	    *(strrchr (path, '/')) = '\0';

	  /* malformed URL */
	  else
	    {
	      g_free (hostname_tmp);
	      g_free (path);
	      g_free (file);

	      *error = UC_ERROR_MALFORMED_LINK;

	      return url_tmp;
	    }

	  p_last = p1;
	}

      ret = g_strconcat (path, "/", p_last, NULL);
    }
  else
    ret = g_strdup (url);

  g_free (hostname_tmp);
  g_free (url_tmp);
  g_free (path);
  g_free (file);

  return ret;
}


/*
 * return a full and normalized url
 */
gchar *
uc_url_normalize (const gchar * current_host,
		  const gchar * current_path, gchar * url)
{
  gchar *protocol = NULL;
  gchar *h_name = NULL;
  gchar *port = NULL;
  gchar *path = NULL;
  gchar *file = NULL;
  gchar *args = NULL;
  gchar *normalized_url = NULL;
  UCError error;

  protocol = uc_url_get_protocol (url);

  if (!strcmp (protocol, "FILE"))
    {
      g_free (protocol);
      return g_strdup (url);
    }

  uc_url_parse (current_host, current_path, url, &h_name,
		&port, &path, &file, &args, &error);

  if (!strcmp (port, "80"))
    normalized_url = g_strconcat (g_strdown (protocol), "://",
				  h_name,
				  (path[0] != '/'
				   && file[0] != '/') ? "/" : "", path,
				  file, args, NULL);
  else
    normalized_url = g_strconcat (g_strdown (protocol), "://",
				  h_name, ":", port,
				  (path[0] != '/'
				   && file[0] != '/') ? "/" : "", path,
				  file, args, NULL);

  g_free (protocol);
  g_free (h_name);
  g_free (port);
  g_free (path);
  g_free (file);
  g_free (args);

  return normalized_url;
}


/*
 * add the given protocol to the given url
 */
gchar *
uc_url_add_protocol (const gchar * proto, const gchar * host)
{
  if ((strstr (host, "://")))
    return g_strdup (host);

  return g_strconcat (proto, "://", host, NULL);
}


/*
 * get the hostname of a given url.
 * if the url doesn't have host name, then
 * current_host is send
 */
gchar *
uc_url_get_hostname (const gchar * current_host, const gchar * url)
{
  gchar **split_proto = NULL;
  gchar **split_slash = NULL;
  gchar *tmp = NULL;

  if ((!url) || (!strlen (url)) || url[0] == '/')
    return (current_host) ? g_strdup (current_host) : NULL;

  split_proto = g_strsplit (url, "://", 2);
  if (!split_proto[1])
    {
      g_strfreev (split_proto);
      return (current_host) ? g_strdup (current_host) : NULL;
    }

  split_slash = g_strsplit (split_proto[1], "/", 2);

  if (!strchr (split_slash[0], ':'))
    tmp = g_strdup (split_slash[0]);
  else
    {
      gchar **split_port = NULL;

      split_port = g_strsplit (split_slash[0], ":", 2);
      tmp = g_strdup (split_port[0]);
      g_strfreev (split_port);
    }

  g_strfreev (split_proto);
  g_strfreev (split_slash);

  return tmp;
}


/*
 * Update the path and file fields of the UCLinkProperties
 * structure -> if there is a "Last-Modified" field in the
 * header, we surely have to deal with a file
 */
void
uc_url_correction (UCLinkProperties * prop)
{
  gchar *begin = NULL;
  gchar *tmp = NULL;

  if (!prop || (strlen (prop->file)) || (strlen (prop->path) <= 1)
      || (!strchr (prop->path, '/'))
      || (prop->path[strlen (prop->path) - 1] == '/')
      || !g_hash_table_lookup (prop->header, "Last-Modified"))
    return;

  begin = strrchr (prop->path, '/');
  tmp = uc_utils_strdup_delim (begin + 1, &(prop->path[strlen (prop->path)]));
  g_free (prop->file);
  prop->file = g_strdup (tmp);

  g_free (tmp);
  tmp = uc_utils_strdup_delim (&prop->path[0], begin + 1);
  g_free (prop->path);
  prop->path = g_strdup (tmp);
  g_free (tmp);
}


/*
 * parse a given url
 */
void
uc_url_parse (const gchar * current_host,
	      const gchar * current_path, gchar * url, gchar ** host,
	      gchar ** port, gchar ** dir, gchar ** file, gchar ** args,
	      UCError * error)
{
  const gchar *p = NULL;
  const gchar *host_b = NULL, *host_e = NULL;
  const gchar *path_b = NULL, *path_e = NULL;
  const gchar *port_b = NULL, *port_e = NULL;
  const gchar *last_slash = NULL;
  gchar *path = NULL;
  gchar *str = NULL;
  gchar *url_tmp = NULL;
  gboolean have_scheme = FALSE;

  url_tmp = uc_utils_replace (url, "&amp;", "&");

  if (strchr (url_tmp, '#'))
    {
      gchar **split = NULL;

      split = g_strsplit (url_tmp, "#", 2);
      str = g_strdup (split[0]);

      g_strfreev (split);
    }
  else
    str = g_strdup (url_tmp);

  if (strlen (str))
    {

      if ((p = strstr (str, "://")))
	{
	  p += 3;
	  have_scheme = TRUE;

	  host_b = p;
	  p = uc_utils_strpbrk_or_eos (p, ":/;?#");
	  host_e = p;
	  *host = uc_utils_strdup_delim (host_b, host_e);

	  if (*p == ':')
	    {
	      port_b = ++p;
	      p = uc_utils_strpbrk_or_eos (p, "/;?#");
	      port_e = p;
	      *port = uc_utils_strdup_delim (port_b, port_e);
	    }
	}
      else
	{
	  p = str;
	  *host = g_strdup (current_host);
	}

      if (strchr (p, ';'))
	{
	  gchar **split = NULL;

	  split = g_strsplit (p, ";", 2);
	  *args = g_strconcat (";", split[1], NULL);

	  g_strfreev (split);
	  ((gchar *) p)[strcspn (p, ";")] = '\0';
	}
      else if (strchr (p, '?'))
	{
	  gchar **split = NULL;

	  split = g_strsplit (p, "?", 2);
	  *args = g_strconcat ("?", split[1], NULL);

	  g_strfreev (split);
	  ((gchar *) p)[strcspn (p, "?")] = '\0';
	}
      else
	*args = g_strdup ("");

      if (!port_b)
	*port = g_strdup (URL_DEFAULT_PORT);

      if (*p == '/')
	{
	  path_b = p;
	  p = uc_utils_strpbrk_or_eos (p, ";?#");
	  path_e = p;
	}
      else
	path_b = path_e = p;

      path = uc_utils_strdup_delim (path_b, path_e);

      last_slash = strrchr (path, '/');
      if (!last_slash)
	{
	  if (!have_scheme)
	    {
	      *dir = g_strdup (current_path);
	      *file = (!have_scheme) ? g_strdup (str) : g_strdup (path);
	    }
	  else
	    {
	      *dir = g_strdup ("");
	      *file = g_strdup ("");
	    }
	}
      else
	{
	  *dir = uc_utils_strdup_delim (path, last_slash);
	  *file = g_strdup (last_slash);
	}

      if (!have_scheme)
	{
	  if ((!strlen (*dir)) && (!strlen (*file)))
	    {
	      if (strchr (*host, '.'))
		{
		  *file = g_realloc (*file, strlen (*host) + 1);
		  strcat (*file, *host);
		}
	      else
		{
		  g_free (*dir);
		  *dir = g_strdup (*host);
		}

	      g_free (*host);
	      *host = g_strdup (current_host);
	    }

	  if (g_strcasecmp (current_host, *host))
	    {
	      gchar *tmp = NULL;

	      tmp = g_strconcat (*host, *dir, NULL);
	      g_free (*host);
	      g_free (*dir);

	      *host = g_strdup (current_host);
	      *dir = tmp;
	    }
	}

      if (!strlen (*dir) && (strlen (*file)) && (!strchr (*file, '/')))
	{
	  g_free (*dir);
	  if (strlen (current_path))
	    *dir = g_strdup (current_path);
	  else
	    *dir = g_strdup ("/");
	}

      if ((!strlen (*dir)) &&
	  (!strlen (*file)) && ((*host)[strlen (*host) - 1] != '/'))
	{
	  g_free (*dir);
	  *dir = g_strdup ("/");
	}

      if ((strlen (*dir)) && ((*dir)[strlen (*dir) - 1] != '/')
	  && ((*file)[0] != '/'))
	{
	  gchar *tmp = NULL;

	  tmp = g_strconcat (*dir, "/", NULL);
	  g_free (*dir);
	  *dir = tmp;
	}

      if (!strchr (*file, '.'))
	{
	  gchar *tmp = NULL;

	  tmp = g_strconcat ((*dir) ? *dir : "", *file, NULL);
	  g_free (*dir);
	  g_free (*file);

	  *file = g_strdup ("");
	  *dir = tmp;
	}
    }
  else
    {
      *host = g_strdup (current_host);
      *dir = g_strdup ("/");
      *file = g_strdup ("");
      *port = g_strdup (URL_DEFAULT_PORT);
      *args = g_strdup ("");
    }

  if (((*dir)[strlen (*dir) - 1] != '/') && (strlen (*file) > 1)
      && ((*file)[0] != '/'))
    {
      g_free (*dir);
      *dir = g_strdup ("/");
    }

  g_free (url_tmp);
  g_free (str);
  g_free (path);

  if (((*dir)[strlen (*dir) - 1] != '/') &&
      (strlen (*file)) && (*file[0] != '/'))
    {
      gchar *tmp = NULL;

      tmp = g_strconcat ("/", *file, NULL);
      g_free (*file);
      *file = tmp;
    }

  /* encode file */
  str = uc_url_encode (*file);
  g_free (*file);
  *file = str;

  /* encode path */
  path = uc_url_encode (*dir);
  g_free (*dir);
  *dir = path;

  if (!strcmp (*port, "80") && !strcasecmp (*host,
					    uc_project_get_current_host ()))
    {
      g_free (*port);
      *port = g_strdup (uc_project_get_current_port ());
    }

  /* if there is relative path indications in the url
   * we should apply a special treatment*/
  if (strstr (*dir, "/.") || strstr (*dir, "./") || strstr (*file, "./"))
    {
      gchar *relative_url = NULL;
      gchar *absolute_url = NULL;

      relative_url = g_strconcat ("http://", *host,
				  (*dir[0] != '/'
				   && *file[0] != '/') ? "/" : "", *dir,
				  *file, *args, NULL);
      absolute_url = uc_url_get_absolute (relative_url, &(*error));

      if (*error == UC_ERROR_NONE)
	uc_url_parse (current_host, current_path, absolute_url,
		      &(*host), &(*port), &(*dir), &(*file), &(*args),
		      &(*error));

      g_free (relative_url);
      g_free (absolute_url);
    }
}


/*
 * get the port of a given url
 */
gchar *
uc_url_get_port (const gchar * url)
{
  gchar *p1 = NULL;
  gchar *p2 = NULL;

  if (!(p1 = strstr (url, "://")))
    return g_strdup ("80");
  p1 += 3;
  if (!(p2 = strchr (p1, ':')))
    return g_strdup ("80");
  p2++;

  p1 = p2;
  while (isdigit ((int) *p2))
    p2++;

  return uc_utils_strdup_delim (p1, p2);
}


/*
 * get the protocol of a given url
 */
gchar *
uc_url_get_protocol (const gchar * url)
{
  gchar **split = NULL;
  gchar *tmp = NULL;

  if (!strchr (url, ':'))
    return g_strup (g_strdup (PROTOCOL_HTTP));

  split = g_strsplit (url, ":", 0);
  tmp = g_strup (g_strdup (split[0]));
  g_strfreev (split);

  return tmp;
}


/*
 * check the validity of a given url
 */
gboolean
uc_url_is_valid (const gchar * url)
{
  gchar *test = NULL;
  gchar *host = NULL;

  if ((!url) || (!strlen (url))
      || (!(host = uc_url_get_hostname (NULL, url))))
    return FALSE;

  if (!(test = uc_utils_get_ip (host)))
    {
      g_free (host);
      return FALSE;
    }

  g_free (host);
  g_free (test);

  return TRUE;
}
