/*
 * Copyright (C) 2002, 2003 Emmanuel Saracco <esaracco@noos.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "utils.h"


/*
 * display debug informations
 */
void
uc_utils_debug (const gchar * format, ...)
{
  va_list ap;
  gchar *s = NULL;

  if (!uc_project_get_debug_mode ())
    return;

  s = (gchar *) malloc (BUFFER_LEN + 1);
  va_start (ap, format);
  vsnprintf (s, BUFFER_LEN, format, ap);
  va_end (ap);

  g_print ("DEBUG: ");
  g_print (s);
}


/*
 * dump the link properties structure to the standard
 * output
 */
void
uc_utils_debug_dump_link_properties (const UCLinkProperties * prop)
{
  if (!uc_project_get_dump_properties ())
    return;

  g_print ("\n\n"
	   "        link_type: %u\n"
	   "        link_value: %s\n"
	   "        url: %s\n"
	   "        label: %s\n"
	   "        protocol: %s\n"
	   "        h_name: %s\n"
	   "        port: %s\n"
	   "        path: %s\n"
	   "        file: %s\n"
	   "        args: %s\n"
	   "        domain: %s\n"
	   "        header_size: %u\n"
	   "        depth_level: %u\n"
	   "        is_parsable: %u\n"
	   "        is_downloadable: %u\n"
	   "        checked: %u\n",
	   prop->link_type,
	   prop->link_value,
	   prop->url,
	   prop->label,
	   prop->protocol,
	   prop->h_name,
	   prop->port,
	   prop->path,
	   prop->file,
	   prop->args,
	   prop->domain,
	   prop->header_size,
	   prop->depth_level, prop->is_parsable, prop->is_downloadable,
	   prop->checked);
}


/*
 * set a treeview column sortable and 
 * resizable
 */
void
uc_utils_set_userfriendly_treeview_column (const GtkTreeView * tv,
					   const gint position)
{
  GtkTreeViewColumn *column = NULL;

  if (!(column = gtk_tree_view_get_column (GTK_TREE_VIEW (tv), position)))
    return;

  gtk_tree_view_column_set_sort_column_id (column, position);
  gtk_tree_view_column_set_resizable (column, TRUE);
}


/*
 * return the content of a given file
 */
gchar *
uc_utils_get_file_content (const gchar * path)
{
  gchar *buffer = NULL;
  GError *error = NULL;
  gsize length = 0;

  g_file_get_contents (path, &buffer, &length, &error);

  return buffer;
}


/*
 * check email syntax and MX.
 *
 * return codes are:
 *
 * - 0: bad syntax (UC_EMAIL_SYNTAX_BAD)
 * - 1: syntax ok but MX not respond (UC_EMAIL_SYNTAX_MX_BAD)
 * - 2: syntax ok and MX ok (UC_EMAIL_OK)
 */
UCEmailStatus
uc_utils_email_is_valid (const gchar * email, const gboolean check_mx)
{
  gchar *tmp = NULL;
  gchar *user = NULL;
  gchar *domain = NULL;
  gchar *email_tmp = NULL;
  gchar *mx = NULL;
  guint ret = UC_EMAIL_SYNTAX_BAD;

  uc_utils_split_email (email, &user, &domain);
  if (!strlen (user) || !strlen (domain))
    {
      g_free (user);
      g_free (domain);

      return UC_EMAIL_SYNTAX_BAD;
    }

  email_tmp = g_strconcat (user, "@", domain, NULL);

  /* fast syntax control */
  if ((!email_tmp) ||
      (!strlen (email_tmp)) ||
      (strchr ("@.", email_tmp[0])) ||
      (strchr ("@.", email_tmp[strlen (email_tmp) - 1])) ||
      (strchr (email_tmp, ' ')) || (!(tmp = strchr (email_tmp, '@')))
      || (!(tmp = strchr (email_tmp, '.'))))
    {
      g_free (email_tmp);

      return UC_EMAIL_SYNTAX_BAD;
    }

  /* check if MX is alive and listening on
   * port 25 */
  if (check_mx)
    {
      mx = uc_utils_get_mx (domain);

      ret = (uc_utils_mx_is_valid (mx)) ? UC_EMAIL_OK :
	UC_EMAIL_SYNTAX_MX_BAD;

      g_free (mx);
    }
  else
    ret = UC_EMAIL_OK;

  g_free (user);
  g_free (domain);

  return ret;
}


/*
 * fill the two parts of a email address.
 */
void
uc_utils_split_email (const gchar * email, gchar ** user, gchar ** domain)
{
  gchar **components = NULL;
  gchar *sep[] = {
    "@",
    "AT"
  };
  guint i = 0;

  if (!email || (!strchr (email, '@') && !strstr (email, "AT")))
    {
      *user = g_strdup ("");
      *domain = g_strdup ("");

      return;
    }

  i = (strchr (email, '@')) ? 0 : 1;

  components = g_strsplit (email, sep[i], 0);
  *user = g_strdup (components[0]);
  *domain = g_strdup (components[1]);

  g_strfreev (components);
}


/*
 * return the prefered MX for a given
 * domain
 *
 * -> function adapted from 
 *    http://www.sslug.dk/emailarkiv/cprog/1999_10/msg00056.html
 */
gchar *
uc_utils_get_mx (const gchar * domain)
{
  struct rrecord
  {
    gint16 r_type;
    gint16 r_class;
    guint32 r_ttl;
    gint16 r_length;
  };

  static gchar out[1000];
  gchar query[1000];
  HEADER *h;
  gchar *p;
  gint qcount, acount;
  gchar buf[1000];
  gint best;
  struct rrecord *r;

  /* ask libc to do the query:
   * look up 'domain'
   * class C_IN (internet)
   * type  T_MX (mail exchanger record)
   * record results in query, max len 1000
   */
  if (res_search (domain, C_IN, T_MX, query, 1000) < 0)
    return g_strdup (domain);

  h = (HEADER *) query;

  /* skip over queries in packet */
  p = query + sizeof (HEADER);
  qcount = ntohs (h->qdcount);	/* # of queries present */
  acount = ntohs (h->ancount);	/* # of answers present */
  /* get past queries */
  while (qcount--)
    {
      p += dn_expand (query, query + 1000, p, buf, 1000);
      p += 4;			/* qclass, qtype */
    }

  /* now we are in the answers section. Just get the best exchanger,
   * we don't want to spend too much time messing around
   */
  best = 99999;
  out[0] = 0;
  while (acount--)
    {
      /* name that this answer refers to */
      p += dn_expand (query, query + 1000, p, buf, 1000);

      r = (struct rrecord *) p;
      if (htons (r->r_type) == T_MX)
	{			/* MX record */
	  gint pref;
	  gchar *p1 = p + 10;

	  pref = ntohs (*(gint16 *) p1);	/* preference */
	  p1 += 2;
	  if (pref < best)
	    {
	      /* get exchanger name */
	      dn_expand (query, query + 1000, p1, out, 1000);
	      best = pref;
	    }
	}
      p = p + 10 + ntohs (r->r_length);	/* skip RR header and data */
    }

  if (out[0])
    return g_strdup (out);
  else
    return g_strdup (domain);

}


/*
 * TRUE if the given MX is alive and listening
 * on port 25
 */
gboolean
uc_utils_mx_is_valid (gchar * mx)
{
  return (uc_connection_open (mx, 25) > 0);
}


/*
 * return ip of the given host. if arg is already ip,
 * return it as is.
 */
gchar *
uc_utils_get_ip (const gchar * host)
{
  struct hostent *serverHostEnt = NULL;
  gchar *buffer = NULL;

  if ((serverHostEnt = gethostbyname (host)) != NULL)
    buffer = g_strdup (inet_ntoa
		       (*((struct in_addr *) serverHostEnt->h_addr_list[0])));
  else if ((in_addr_t) (inet_addr (host)) != (in_addr_t) - 1)
    buffer = g_strdup (host);
  else
    buffer = g_strdup ("");

  return buffer;
}


/*
 * return a formated label to display in the
 * url tree view list
 */
gchar *
uc_utils_format_list_label (gchar * label)
{
  gchar *str = NULL;
  gchar *tmp = NULL;

  if (label)
    uc_utils_replacelr (uc_utils_strstrip (label), ' ');

  if ((!label) || (!strlen (label)))
    tmp = g_strdup ("");
  else if (strlen (label) > ((LIST_LABEL_MAX_LEN - 5) + 1))
    {
      gchar *tmp_cut = NULL;

      tmp_cut = g_memdup (label, (LIST_LABEL_MAX_LEN - 5) + 1);
      tmp_cut[LIST_LABEL_MAX_LEN - 5] = '\0';
      tmp = g_strconcat (tmp_cut, "[...]", NULL);

      g_free (tmp_cut);
    }
  else
    tmp = g_strdup (label);

  if (!g_utf8_validate (tmp, -1, NULL))
    str = g_convert (tmp, -1, "UTF-8", uc_project_get_local_charset (),
		     NULL, NULL, NULL);
  else
    str = g_strdup (tmp);

  g_free (tmp);

  return str;
}


/*
 * replace one or more characters in a given
 * string
 */
gchar *
uc_utils_replace (gchar * str, const gchar * old, const gchar * new)
{
  guint i = 0;
  guint str_len = strlen (str);
  guint old_len = strlen (old);
  GString *tmp = g_string_new ("");
  gchar *ret = NULL;

  for (i = 0; i < str_len; i++)
    if (!memcmp (&str[i], old, old_len))
      tmp = g_string_append (tmp, new), i += old_len - 1;
    else
      tmp = g_string_append_c (tmp, str[i]);

  ret = tmp->str;
  g_string_free (tmp, FALSE);

  return ret;
}


/*
 * replace a caracter by another (do not allocate
 * new memory for the result string)
 */
gchar *
uc_utils_replace1 (gchar * string, const gchar c1, const char c2)
{
  gchar *p = string;

  do
    if (*p == c1)
      *p = c2;
  while (*p++);

  return string;
}


/*
 * remove all line return in a given
 * string
 */
gchar *
uc_utils_replacelr (gchar * string, const gchar c)
{
  gchar *p = string;

  do
    if (*p == '\n' || *p == '\r')
      *p = c;
  while (*p++);

  return string;
}


/*
 * remove all spaces or separators before a
 * given string
 */
gchar *
uc_utils_strchomp (gchar * string)
{
  gsize len;

  len = strlen (string);
  while (len--)
    {
      if ((string[len] == '\n') ||
	  (string[len] == '\r') ||
	  (string[len] == '\t') || (g_ascii_isspace (string[len])))
	string[len] = '\0';
      else
	break;
    }

  return string;
}


/*
 * remove all spaces or separator after a
 * given string
 */
gchar *
uc_utils_strchug (gchar * string)
{
  gchar *start;

  for (start = (gchar *) string;
       (*start) && ((*start == '\n') ||
		    (*start == '\r') ||
		    (*start == '\t') || (g_ascii_isspace (*start))); start++);

  g_memmove (string, start, strlen ((gchar *) start) + 1);

  return string;
}


/*
 * convert a cache uid to filename
 */
gchar *
uc_utils_convert_uid2file (const guint uid)
{
  gchar path[BUFFER_URL_LEN + 1] = { 0 };

  if ((snprintf
       (path, BUFFER_URL_LEN, "%s/%s/%d", uc_project_get_working_path (),
	uc_project_get_cache_name (), uid) < 0))
    return NULL;

  return g_strdup (path);
}


/*
 * fill given args with the gconf values for
 * the gnome network proxy settings
 */
void
uc_utils_get_gnome_proxy_conf (gchar ** host, guint * port)
{
  GConfClient *client = NULL;

  client = gconf_client_get_default ();

  if (gconf_client_get_bool
      (client, "/system/http_proxy/use_http_proxy", NULL))
    {
      *host = gconf_client_get_string (client,
				       "/system/http_proxy/host", NULL);
      if (!
	  (*port =
	   gconf_client_get_int (client, "/system/http_proxy/port", NULL)))
	*port = 8080;
    }

  if (!(*host))
    *host = g_strdup ("");
}


/*
 * true if it exist enough space on the parition
 * of the given path to write on it
 */
gboolean
uc_utils_disk_free_enough (const guint requested_size)
{
  struct statfs sfs;
  guint requested_blocks = 0;
  guint secure_add = UC_SECURE_BYTES_ADDING;

  statfs (uc_project_get_working_path (), &sfs);

  secure_add /= sfs.f_bsize;
  requested_blocks = ((requested_size / sfs.f_bsize) <= 0) ?
    1 : requested_size / sfs.f_bsize;

  return ((requested_blocks + secure_add) < sfs.f_bavail);
}


/*
 * clean a link value string
 */
gchar *
uc_utils_clean_tag_link_value (const gchar * value)
{
  gchar tmp[BUFFER_URL_LEN + 1] = { 0 };
  guint i = 0;
  guint j = 0;

  if (!value)
    return g_strdup ("");

  for (j = 0; j < strlen (value); j++)
    if ((i < BUFFER_URL_LEN)
	&& (value[j] != '\'' && value[j] != '"' && value[j] != '>'))
      tmp[i++] = value[j];

  return g_strdup (tmp);
}


/*
 * return a pointer on the next char after the searched string
 * if it found, else return the original pointer
 */
gpointer
uc_utils_search_string_next (gpointer buf, const gchar * str,
			     const gchar limit_char)
{
  gpointer sav = buf;
  gchar *buffer = (gchar *) buf;

  if ((!str) || (!strlen (str)) || (!buffer)
      || (strlen (buffer) < strlen (str)))
    return (gpointer) buf;

  while (*buffer &&
	 (!uc_utils_memcasecmp (buffer, str)) && (*buffer != limit_char))
    buffer++;

  if (*buffer != limit_char)
    sav = (buffer += strlen (str));

  return sav;
}


/*
 * comparison between two strings, based on the length of
 * the second string
 */
gboolean
uc_utils_memcasecmp (const gchar * str1, const gchar * str2)
{
  gchar tmp[BUFFER_URL_LEN + 1] = { 0 };
  guint len = 0;

  if (!str1 || !str2)
    return FALSE;

  len = strlen (str2);

  if ((len > BUFFER_URL_LEN) || (strlen (str1) < len))
    return FALSE;

  if (!memcmp (str1, str2, len))
    return TRUE;

  memcpy (tmp, str1, len);
  return (!g_strcasecmp (tmp, str2));
}


/*
 * make given path -- create intermediate
 * paths if they don't exist -- and test if we are
 * authorized to write in
 */
gboolean
uc_utils_mkdirs (const gchar * path, const gboolean create_all)
{
  gchar test_path[BUFFER_URL_LEN + 1] = { 0 };
  gchar **paths = NULL;
  gboolean ret = FALSE;
  gint fd;
  gint i = 0;

  if (create_all)
    {
      paths = g_strsplit (path, "/", 255);
      while (paths[i])
	{
	  if (strlen (paths[i]))
	    {
	      strncat (test_path, "/", BUFFER_URL_LEN - strlen (test_path));
	      if (!strncat
		  (test_path, paths[i], BUFFER_URL_LEN - strlen (test_path)))
		{
		  g_strfreev (paths);
		  return FALSE;
		}

	      mkdir (test_path, S_IRWXU);
	    }
	  i++;
	}
    }
  else
    {
      mkdir (path, S_IRWXU);
      strncat (test_path, path, BUFFER_URL_LEN - strlen (test_path));
    }

  strncat (test_path, "/filetest", BUFFER_URL_LEN - strlen (test_path));
  if ((fd = open (test_path, O_CREAT, S_IRUSR | S_IWUSR)) >= 0)
    {
      close (fd);
      unlink (test_path);
      ret = TRUE;
    }

  g_strfreev (paths);
  return ret;
}


/*
 * remove all files in a given
 * directory
 */
void
uc_utils_rmfiles (const gchar * path)
{
  DIR *dir = NULL;
  struct dirent *file = NULL;

  if ((dir = opendir (path)))
    {
      while ((file = readdir (dir)))
	{
	  gchar *file_path = g_strconcat (path, "/", file->d_name,
					  NULL);
	  unlink (file_path);
	  g_free (file_path);
	}
      closedir (dir);
    }
}


/*
 * remove given path -- remove intermediate
 * paths too
 */
void
uc_utils_rmdirs (const gchar * path, const gboolean delete_all)
{
  gchar **paths = NULL;
  gint i = 0;

  if (delete_all)
    {
      paths = g_strsplit (path, "/", 255);
      i = uc_utils_vector_length (paths);
      while (i >= 0)
	{
	  gchar dir[BUFFER_URL_LEN + 1] = { 0 };
	  guint j = 0;

	  for (j = 0; j < i; j++)
	    if (strlen (paths[j]))
	      {
		if ((!strncat
		     (dir, "/",
		      BUFFER_URL_LEN -
		      strlen (dir)))
		    ||
		    (!strncat (dir, paths[j], BUFFER_URL_LEN - strlen (dir))))
		  {
		    g_strfreev (paths);
		    return;
		  }
	      }

	  uc_utils_rmfiles (dir);
	  rmdir (dir);

	  i--;
	}

      g_strfreev (paths);
    }
  else
    {
      uc_utils_rmfiles (path);
      rmdir (path);
    }
}


/*
 * return the length of a given vector
 */
guint
uc_utils_vector_length (gpointer data)
{
  guint i = 0;
  gchar **vector = (gchar **) data;

  while (vector[i++]);

  return --i;
}


/*
 * check if the given path is a valid
 * for gURLChecker work 
 */
gboolean
uc_utils_working_path_is_valid (const gchar * path)
{
  gboolean ret = FALSE;
  gchar *final_path = NULL;

  final_path = g_strconcat (path, "/", uc_project_get_cache_name (), NULL);

  if ((ret = uc_utils_mkdirs (final_path, FALSE)))
    uc_utils_rmdirs (final_path, FALSE);

  g_free (final_path);
  return ret;
}

/*
 * check if the given file is a valid
 * executable file
 */
gboolean
uc_utils_file_is_executable (const gchar * file)
{
  struct stat s;

  return ((!stat (file, &s)) && (s.st_mode & S_IXOTH));
}


/*
 * check if a connection if alive or
 * not
 *
 * -> taken from wget source code
 */
guint
uc_utils_test_socket_open (const guint sock)
{
  fd_set check_set;
  struct timeval to;

  /* Check if we still have a valid (non-EOF) connection.  From Andrew
   * Maholski's code in the Unix Socket FAQ.  */

  FD_ZERO (&check_set);
  FD_SET (sock, &check_set);

  /* Wait one microsecond */
  to.tv_sec = 0;
  to.tv_usec = 1;

  /* If we get a timeout, then that means still connected */
  if (select (sock + 1, &check_set, NULL, NULL, &to) == 0)
    {
      /* Connection is valid (not EOF), so continue */
      return 1;
    }
  else
    return 0;
}


/*
 * make a string with 2 given
 * pointers
 */
gchar *
uc_utils_strdup_delim (const gchar * begin, const gchar * end)
{
  gchar *res = NULL;

  res = (gchar *) g_malloc (end - begin + 1);
  memcpy (res, begin, end - begin);
  res[end - begin] = '\0';

  return res;
}


/*
 * return a pointer on a given char address
 * or on the end of the passed string
 */
gchar *
uc_utils_strpbrk_or_eos (const gchar * str, const gchar * accept)
{
  gchar *p = NULL;

  p = strpbrk (str, accept);
  if (!p)
    p = (gchar *) str + strlen (str);

  return p;
}


/*
 * return TRUE if the first argument is in
 * the second argument vector
 */
gboolean
uc_utils_array_is_in (const gchar * item, gpointer array, const guint size)
{
  gchar **items = (gchar **) array;
  guint i = 0;

  for (i = 0; i < size; i++)
    if (!strcmp (item, items[i]))
      return TRUE;

  return FALSE;
}
