/* 
 * Copyright (C) 2002, 2003 Emmanuel Saracco <esaracco@noos.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA. 
 */

#include "web_site.h"


static GList *uc_web_site_get_links_real (UCLinkProperties * url,
					  const gchar * host,
					  const guint depth);
static void
uc_web_site_get_links_real_clean (gchar ** current_host,
				  gchar ** current_port,
				  gchar ** current_path,
				  gchar ** current_file,
				  gchar ** current_args);

static GList *uc_check_url_get_links (UCLinkProperties * url,
				      const guint depth);


/*
 * free all allocated variable from
 * uc_web_site_get_links_real* like functions
 */
static void
uc_web_site_get_links_real_clean (gchar ** current_host,
				  gchar ** current_port,
				  gchar ** current_path,
				  gchar ** current_file,
				  gchar ** current_args)
{
  if (current_host)
    g_free (*current_host);
  if (current_port)
    g_free (*current_port);
  if (current_path)
    g_free (*current_path);
  if (current_file)
    g_free (*current_file);
  if (current_args)
    g_free (*current_args);
}


/*
 * get all the links of a page
 */
static GList *
uc_web_site_get_links_real (UCLinkProperties * prop,
			    const gchar * host, const guint depth)
{
  GList *list = NULL;
  UCLinkProperties *lp = NULL;
  gchar *current_host = NULL;
  gchar *current_path = NULL;
  gchar *current_port = NULL;
  gchar *current_file = NULL;
  gchar *current_args = NULL;
  GList *tags = NULL;
  GList *item = NULL;
  gchar *normalized_url = NULL;
  UCError error = UC_ERROR_NONE;
  gboolean already_checked = FALSE;
  gboolean accept = FALSE;

  /* current informations for timeouts
   * callback */
  currentItem.parent = (depth) ? prop : NULL;

  uc_url_parse (prop->h_name, prop->path, prop->url, &current_host,
		&current_port, &current_path, &current_file, &current_args,
		&error);

  /* register the main project url in the
   * checked links */
  if (!depth)
    {
      normalized_url =
	uc_url_normalize (current_host, current_path, uc_project_get_url ());
      (gboolean) uc_check_link_already_checked_with_insert
	(prop, normalized_url);
      list = g_list_prepend (list, (gpointer) prop);
      g_free (normalized_url);
    }

  if ((tags = prop->tags))
    {
      item = tags;
      while (UC_CHECK_PARSER_WHILE_CONDITION)
	{
	  UCHTMLTag *tag = (UCHTMLTag *) item->data;

	  uc_application_update_ui ();

	  uc_check_ignore_item_set_value (FALSE);

	  /* if user have requested to suspend the check for
	   * a while */
	  while (UC_CHECK_PARSER_SUSPEND_CONDITION)
	    uc_application_update_ui ();

	  item = g_list_next (item);

	  /* META tag (already processed in the 
	   * uc_check_link_get_properties () function ) */
	  if (tag->type == LINK_TYPE_META || tag->type == LINK_TYPE_NONE)
	    continue;

	  uc_application_display_state_message (CHECK_MESSAGE_LABEL_SECOND,
						tag->value);
	  uc_application_update_ui ();

	  uc_report_set_alllinks (0);

	  /* if link value is malformed, do not treat it as
	   * normal */
	  if (!tag->malformed)
	    {
	      normalized_url =
		uc_url_normalize (current_host, current_path, tag->value);
	      already_checked =
		uc_check_link_already_checked_with_insert
		(prop, normalized_url);
	    }
	  else
	    {
	      uc_check_insert_malformed_link (tag, list);
	      continue;
	    }

	  if (!already_checked)
	    {
	      uc_check_wait ();

	      uc_check_currentitem_init (normalized_url, tag->label, NULL);

	      /* retrieve properties of the link and check if its
	       * ok for register */
	      if ((lp =
		   uc_check_link_get_properties (depth,
						 current_host,
						 current_path,
						 (UCHTMLTag *)
						 tag, &accept)) && accept)
		list = g_list_append (list,
				      uc_check_register_link (normalized_url,
							      lp));
	      /* link rejected */
	      else
		uc_lists_checked_links_node_free ((gpointer) lp);
	    }

	  g_free (normalized_url);
	  uc_report_display_update ();
	}
    }

  /* add timeout links list to the main
   * urls list */
  uc_check_register_timeout_links (&list);
  uc_check_timeout_links_free ();

  uc_web_site_get_links_real_clean (&current_host,
				    &current_port,
				    &current_path, &current_file,
				    &current_args);

  return list;
}


/*
 * put in a list all the urls of
 * a given url
 */
static GList *
uc_check_url_get_links (UCLinkProperties * prop, const guint depth)
{
  gchar *host = NULL;
  GList *urls = NULL;
  GList *item = NULL;
  GList *tmp_list = NULL;

  if (!prop)
    return NULL;

  host = uc_url_get_hostname ("", prop->url);

  if (UC_CHECK_ABORT_IF_CONDITION)
    {
      g_free (host);
      return NULL;
    }

  uc_application_display_state_message (CHECK_MESSAGE_LABEL_FIRST, prop->url);
  uc_application_display_state_message (CHECK_MESSAGE_LABEL_SECOND, "");
  uc_application_display_state_message (CHECK_MESSAGE_LABEL_THIRD,
					_("Searching..."));
  uc_application_update_ui ();

  urls = uc_web_site_get_links_real (prop, host, depth);
  item = urls;
  while (item)
    {
      UCLinkProperties *lp = (UCLinkProperties *) item->data;

      uc_application_update_ui ();

      if (lp->is_parsable && !lp->checked)
	{
	  /* mark this link as already checked */
	  prop->checked = TRUE;

	  uc_utils_debug ("Checking %s...\n", lp->url);
	  lp->childs = uc_check_url_get_links (lp, depth + 1);
	}

      tmp_list = g_list_append (tmp_list, (gpointer) lp);

      item = g_list_next (item);
    }

  g_free (host);
  g_list_free (urls);

  return tmp_list;
}


/*
 * begin the page urls check
 */
void
uc_web_site_begin_check (void)
{
  time_t t;
  UCLinkProperties *prop = NULL;
  gchar *tmp = NULL;
  GtkWidget *widget = NULL;
  gboolean accept = FALSE;
  guint32 timer_id = 0;
  guint i_iter = 0;
  GtkTreeIter iter[UC_MAX_DEPTH_LEVEL];
  UCHTMLTag tag;

  uc_application_draw_main_frames ();
  uc_application_update_ui ();

  UC_GET_WIDGET ("urls_list", main_window, widget);
  treeview = (GtkTreeView *) widget;

  uc_application_progress_dialog_show ();
  uc_application_update_ui ();

  currentItem.url = uc_project_get_url ();
  uc_project_set_current_host (uc_url_get_hostname (NULL,
						    uc_project_get_url ()));
  uc_project_set_current_port (uc_url_get_port (uc_project_get_url ()));
  tmp = g_strdup_printf (_("Checking %s..."), uc_project_get_url ());
  uc_application_display_state_message (CHECK_MESSAGE_LABEL_FIRST, tmp);
  uc_application_update_ui ();
  g_free (tmp);

  uc_check_reset ();
  uc_project_free (UC_PROJECT_TYPE_WEB_SITE);
  tmp = g_strconcat ("gURLChecker v", VERSION, " - ", uc_project_get_url (),
		     NULL);
  gtk_window_set_title (GTK_WINDOW (main_window), tmp);
  g_free (tmp);

  time (&t);
  timer_id = gtk_timeout_add (1000, &uc_report_timer_callback, (gpointer) t);
  tag.name = g_strdup ("");
  tag.label = g_strdup ("");
  tag.value = g_strdup (uc_project_get_url ());
  tag.attr = NULL;
  tag.type = LINK_TYPE_HREF;
  tag.malformed = FALSE;

  if (!
      (prop =
       uc_check_link_get_properties (0, uc_project_get_current_host (), "",
				     (UCHTMLTag *) & tag, &accept)))
    uc_application_dialog_show (_
				("The specified project URL does not "
				 "respond.\n"
				 "Check has been canceled."),
				GTK_MESSAGE_ERROR);
  else
    {
      prop->is_parsable = TRUE;

      /* if there was a redirection for the main
       * url */
      if (strcasecmp (prop->h_name, uc_project_get_current_host ()))
	{
	  uc_project_set_current_host (g_strdup (prop->h_name));
	}

      uc_utils_debug ("Checking %s...\n", prop->url);

      uc_project_set_check_is_current (TRUE);
      uc_check_register_link (prop->url, prop);
      uc_lists_checked_links_set (uc_check_url_get_links (prop, 0));
      uc_project_set_check_is_current (FALSE);
    }
  gtk_timeout_remove (timer_id);

  i_iter = 0;
  treestore = gtk_tree_store_new (N_COLUMNS,
				  G_TYPE_INT,
				  GDK_TYPE_PIXBUF,
				  GDK_TYPE_PIXBUF,
				  G_TYPE_STRING,
				  G_TYPE_STRING, GDK_TYPE_COLOR);

  uc_check_display_list (uc_lists_checked_links_get (), iter, i_iter, 0);

  uc_application_build_url_treeview ();

  gtk_widget_destroy (progress_dialog);

  /* activate some menu items */
  uc_application_menu_set_sensitive_all ("display", TRUE);
  uc_application_menu_set_sensitive_all ("search", TRUE);

  /* treeview is collapsed by default and no row is selected, so we deactivate 
     the "Collapse all the list" menu options */
  UC_WIDGET_SET_SENSITIVE (glade_main_window, "display_all", FALSE);
  UC_WIDGET_SET_SENSITIVE (glade_main_window, "display_collapse_all", FALSE);

  uc_check_cancel_set_value (FALSE);
}
