/*
 * Copyright (C) 2002, 2003 Emmanuel Saracco <esaracco@noos.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "xml_parser.h"


static UCLinkProperties *parents[UC_MAX_DEPTH_LEVEL];

static xmlDocPtr uc_xml_parser_parse_file (const gchar * docname);
static GList *uc_xml_parser_parse_bookmark (guint depth, GList * list,
					    xmlDocPtr doc, xmlNodePtr cur);
static GList *uc_xml_parser_parse_folder (guint depth, GList * list,
					  xmlDocPtr doc, xmlNodePtr cur);


/*
 * parse bookmark entity
 */
static GList *
uc_xml_parser_parse_bookmark (guint depth,
			      GList * list, xmlDocPtr doc, xmlNodePtr cur)
{
  xmlChar *key = NULL;
  xmlChar *value = NULL;
  UCHTMLTag *tag = NULL;
  xmlNodePtr parent = NULL;

  uc_report_display_update ();

  /* if user have requested to suspend the check for
   * a while */
  while (UC_CHECK_PARSER_SUSPEND_CONDITION)
    uc_application_update_ui ();

  uc_check_wait ();

  value = xmlGetProp (cur, (const xmlChar *) "href");

  parent = cur;
  cur = cur->xmlChildrenNode;
  while (cur && !uc_check_cancel_get_value ())
    {
      if (!xmlStrcmp (cur->name, (const xmlChar *) "title"))
	{
	  UCLinkProperties *prop = NULL;
	  gboolean accept;

	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);

	  tag = uc_html_parser_node_new (NULL);
	  tag->depth = depth;
	  tag->name = g_strdup ("A");
	  tag->label = g_strdup (key);
	  tag->value = g_strdup (value);
	  tag->type = LINK_TYPE_HREF;

	  uc_check_currentitem_init (value, key, &parent);

	  if ((prop =
	       uc_check_link_get_properties (depth, "", "", tag, &accept))
	      && accept)
	    {
	      prop->xml_node = parent;
	      list =
		g_list_append (list,
			       uc_check_register_link (prop->url, prop));
	    }
	  else
	    uc_lists_checked_links_node_free (prop);

	  g_free (tag->name);
	  g_free (tag->label);
	  g_free (tag->value);
	  g_free (tag);

	  xmlFree (key);

	  break;
	}

      cur = cur->next;
    }

  xmlFree (value);

  return list;
}


/*
 * parse folder entity
 */
static GList *
uc_xml_parser_parse_folder (guint depth,
			    GList * list, xmlDocPtr doc, xmlNodePtr cur)
{
  xmlChar *key = NULL;
  xmlNodePtr parent = NULL;

  uc_report_display_update ();

  if (depth == uc_project_get_depth_level ())
    return list;

  parent = cur;
  cur = cur->xmlChildrenNode;
  while (cur && !uc_check_cancel_get_value ())
    {
      if (!xmlStrcmp (cur->name, (const xmlChar *) "title"))
	{
	  UCLinkProperties *prop = NULL;

	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);

	  prop = uc_check_link_properties_node_new ();
	  prop->xml_node = parent;
	  prop->depth_level = depth;
	  prop->label = g_strdup (key);
	  prop->link_value = g_strdup ("");
	  prop->link_icon =
	    gdk_pixbuf_new_from_file (UC_PIXMAPS_DIR "/link_type_folder.png",
				      NULL);
	  prop->link_type = LINK_TYPE_BOOKMARK_FOLDER;
	  prop->header = g_hash_table_new (g_str_hash, g_str_equal);

	  uc_check_link_already_checked_with_insert (prop, "");
	  list = g_list_append (list, uc_check_register_link ("", prop));

	  parents[depth] = prop;

	  xmlFree (key);
	}
      else if (!xmlStrcmp (cur->name, (const xmlChar *) "bookmark"))
	parents[depth]->childs =
	  uc_xml_parser_parse_bookmark (depth + 1, parents[depth]->childs,
					doc, cur);
      else if (!xmlStrcmp (cur->name, (const xmlChar *) "folder"))
	parents[depth]->childs =
	  uc_xml_parser_parse_folder (depth + 1, parents[depth]->childs,
				      doc, cur);

      cur = cur->next;
    }

  if (parents[depth])
    uc_check_register_timeout_links (&parents[depth]->childs);
  uc_check_timeout_links_free ();

  return list;
}


/*
 * open xml file and parse it with DOM.
 *
 * -> if there is a encoding problem, rebuild the given file by specifying
 * the encoding based on current locales and try to parse it again.
 */
static xmlDocPtr
uc_xml_parser_parse_file (const gchar * docname)
{
  xmlDocPtr doc = NULL;

  if (!(doc = xmlParseFile (docname)))
    {
      FILE *fd = NULL;
      gchar *buffer = NULL;
      gchar **tab = NULL;
      gchar *filename = NULL;
      guint i = 0;

      filename = g_strdup_printf ("%s/%s/%s.%u",
				  uc_project_get_working_path (),
				  uc_project_get_cache_name (),
				  strrchr (docname, '/'), getpid ());
      fd = fopen (filename, "w");

      g_file_get_contents (docname, &buffer, NULL, NULL);
      tab = g_strsplit (buffer, "\n", 0);

      while (tab[i])
	{
	  if (strstr (tab[i], "<?xml "))
	    {
	      g_free (tab[i]);
	      tab[i] =
		g_strdup_printf ("<?xml version=\"1.0\" encoding=\"%s\"?>",
				 uc_project_get_local_charset ());

	      break;
	    }
	  i++;
	}

      if (!tab[i])
	fprintf (fd,
		 "<?xml version=\"1.0\" encoding=\"%s\"?>\n"
		 "%s", uc_project_get_local_charset (), buffer);
      else
	{
	  g_free (buffer);
	  buffer = g_strjoinv ("\n", tab);
	  fprintf (fd, "%s", buffer);
	}

      fclose (fd);

      doc = xmlParseFile (filename);
      unlink (filename);

      g_free (filename);
      g_strfreev (tab);
      g_free (buffer);
    }

  return doc;
}


/*
 * return the list of checked urls.
 *
 * -> the second argument (xmlDocPtr doc) is allocated. caller must
 * free it with xmlFreeDoc ().
 */
GList *
uc_xml_parser_get_tags (const gchar * docname)
{
  GList *list = NULL;
  xmlNodePtr cur = NULL;

  if (!(doc = uc_xml_parser_parse_file (docname)))
    {
      uc_application_dialog_show (_("The given XML file is not well formed "
				    "or its encoding can not "
				    "be found..\n"
				    "Check has been aborted."),
				  GTK_MESSAGE_ERROR);

      return NULL;
    }

  if (!(cur = xmlDocGetRootElement (doc)))
    {
      uc_application_dialog_show (_("The given XML file is empty.\n"),
				  GTK_MESSAGE_WARNING);

      return NULL;
    }

  cur = cur->xmlChildrenNode;
  while (cur && !uc_check_cancel_get_value ())
    {
      if (!xmlStrcmp (cur->name, (const xmlChar *) "bookmark"))
	list = uc_xml_parser_parse_bookmark (0, list, doc, cur);
      else if (!xmlStrcmp (cur->name, (const xmlChar *) "folder"))
	list = uc_xml_parser_parse_folder (0, list, doc, cur);

      cur = cur->next;
    }

  if (parents[0])
    uc_check_register_timeout_links (&parents[0]->childs);
  else
    uc_check_register_timeout_links (&list);
  uc_check_timeout_links_free ();

  return list;
}
